# Copyright (C) 2024-2025 The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

from functools import partial
from typing import Iterator

import pytest
from pytest_postgresql import factories

from swh.core.db.db_utils import initialize_database_for_module
from swh.storage.proxies.blocking.db import BlockingAdmin, BlockingQuery

blocking_db_postgresql_proc = factories.postgresql_proc(
    load=[
        partial(
            initialize_database_for_module,
            modname="storage.proxies.blocking",
            version=BlockingAdmin.current_version,
        ),
    ],
)


blocking_db_postgresql = factories.postgresql(
    "blocking_db_postgresql_proc",
)


@pytest.fixture
def blocking_admin(blocking_db_postgresql) -> Iterator[BlockingAdmin]:
    with BlockingAdmin.connect(blocking_db_postgresql.info.dsn) as db:
        yield db


@pytest.fixture
def blocking_query(blocking_db_postgresql) -> Iterator[BlockingQuery]:
    with BlockingQuery.connect(blocking_db_postgresql.info.dsn) as db:
        yield db



from collections import defaultdict
import datetime
from typing import Set, Tuple

import pytest

from swh.model.model import OriginVisitStatus
from swh.model.swhids import ExtendedSWHID
from swh.storage import get_storage
from swh.storage.interface import StorageInterface


@pytest.fixture
def swh_storage(mocker):
    storage = get_storage(
        cls="pipeline",
        steps=[
            {"cls": "record_references"},
            {"cls": "memory"},
        ],
    )

    storage.object_references_create_partition(
        *datetime.date.today().isocalendar()[0:2]
    )

    return mocker.Mock(wraps=storage)


def check_references_recorded(
    storage: StorageInterface, references: Set[Tuple[str, str]]
) -> None:
    expected = defaultdict(set)

    for source, target in references:
        expected[ExtendedSWHID.from_string(target)].add(
            ExtendedSWHID.from_string(source)
        )

    for target_swhid, expected_sources in expected.items():
        recorded_refs = storage.object_find_recent_references(
            target_swhid, limit=len(expected_sources) + 1
        )

        assert set(recorded_refs) == expected_sources


def test_record_references_adding_content(swh_storage, sample_data):
    stats = swh_storage.content_add(sample_data.contents)
    assert "object_reference:add" not in stats
    swh_storage.object_references_add.assert_not_called()


def test_record_references_adding_directories(swh_storage, sample_data):
    stats = swh_storage.directory_add(
        [
            sample_data.directory,
            sample_data.directory2,
            sample_data.directory3,
            sample_data.directory5,
        ]
    )

    assert stats["directory:add"] > 0
    assert stats["object_reference:add"] == 6

    check_references_recorded(
        swh_storage,
        {
            # directory → content
            (
                "swh:1:dir:5256e856a0a0898966d6ba14feb4388b8b82d302",
                "swh:1:cnt:d81cc0710eb6cf9efd5b920a8453e1e07157b6cd",
            ),
            # directory → directory5
            (
                "swh:1:dir:5256e856a0a0898966d6ba14feb4388b8b82d302",
                "swh:1:dir:4b825dc642cb6eb9a060e54bf8d69288fbee4904",
            ),
            # directory2 → content2
            (
                "swh:1:dir:8505808532953da7d2581741f01b29c04b1cb9ab",
                "swh:1:cnt:36fade77193cb6d2bd826161a0979d64c28ab4fa",
            ),
            # directory3 → content
            (
                "swh:1:dir:13089e6e544f78df7c9a40a3059050d10dee686a",
                "swh:1:cnt:d81cc0710eb6cf9efd5b920a8453e1e07157b6cd",
            ),
            # directory3 → directory
            (
                "swh:1:dir:13089e6e544f78df7c9a40a3059050d10dee686a",
                "swh:1:dir:5256e856a0a0898966d6ba14feb4388b8b82d302",
            ),
            # directory3 → content2
            (
                "swh:1:dir:13089e6e544f78df7c9a40a3059050d10dee686a",
                "swh:1:cnt:36fade77193cb6d2bd826161a0979d64c28ab4fa",
            ),
        },
    )


def test_record_references_adding_revisions(swh_storage, sample_data):
    stats = swh_storage.revision_add(sample_data.git_revisions)

    assert stats["revision:add"] > 0
    assert stats["object_reference:add"] == 8

    check_references_recorded(
        swh_storage,
        {
            # revision → directory
            (
                "swh:1:rev:01a7114f36fddd5ef2511b2cadda237a68adbb12",
                "swh:1:dir:5256e856a0a0898966d6ba14feb4388b8b82d302",
            ),
            # revision2 → directory2
            (
                "swh:1:rev:a646dd94c912829659b22a1e7e143d2fa5ebde1b",
                "swh:1:dir:8505808532953da7d2581741f01b29c04b1cb9ab",
            ),
            # revision2 → revision
            (
                "swh:1:rev:a646dd94c912829659b22a1e7e143d2fa5ebde1b",
                "swh:1:rev:01a7114f36fddd5ef2511b2cadda237a68adbb12",
            ),
            # revision3 → directory2
            (
                "swh:1:rev:beb2844dff30658e27573cb46eb55980e974b391",
                "swh:1:dir:8505808532953da7d2581741f01b29c04b1cb9ab",
            ),
            # revision3 → revision
            (
                "swh:1:rev:beb2844dff30658e27573cb46eb55980e974b391",
                "swh:1:rev:01a7114f36fddd5ef2511b2cadda237a68adbb12",
            ),
            # revision3 → revision2
            (
                "swh:1:rev:beb2844dff30658e27573cb46eb55980e974b391",
                "swh:1:rev:a646dd94c912829659b22a1e7e143d2fa5ebde1b",
            ),
            # revision4 → directory
            (
                "swh:1:rev:ae860aec43700c7f5a295e2ef47e2ae41b535dfe",
                "swh:1:dir:5256e856a0a0898966d6ba14feb4388b8b82d302",
            ),
            # revision4 → revision3
            (
                "swh:1:rev:ae860aec43700c7f5a295e2ef47e2ae41b535dfe",
                "swh:1:rev:beb2844dff30658e27573cb46eb55980e974b391",
            ),
        },
    )


def test_record_references_adding_releases(swh_storage, sample_data):
    stats = swh_storage.release_add(sample_data.releases)

    assert stats["release:add"] > 0
    assert stats["object_reference:add"] == 3

    check_references_recorded(
        swh_storage,
        {
            # release → revision
            (
                "swh:1:rel:f7f222093a18ec60d781070abec4a630c850b837",
                "swh:1:rev:01a7114f36fddd5ef2511b2cadda237a68adbb12",
            ),
            # release2 → revision2
            (
                "swh:1:rel:db81a26783a3f4a9db07b4759ffc37621f159bb2",
                "swh:1:rev:a646dd94c912829659b22a1e7e143d2fa5ebde1b",
            ),
            # release3 → revision3
            (
                "swh:1:rel:1c5d42e603ce2eea44917fadca76c78bad76aeb9",
                "swh:1:rev:beb2844dff30658e27573cb46eb55980e974b391",
            ),
        },
    )


def test_record_references_adding_snapshots(swh_storage, sample_data):
    stats = swh_storage.snapshot_add(sample_data.snapshots)
    assert stats["snapshot:add"] > 0
    assert stats["object_reference:add"] == 7

    check_references_recorded(
        swh_storage,
        {
            # snapshot → revision
            (
                "swh:1:snp:9b922e6d8d5b803c1582aabe5525b7b91150788e",
                "swh:1:rev:01a7114f36fddd5ef2511b2cadda237a68adbb12",
            ),
            # complete_snapshot → directory
            (
                "swh:1:snp:db99fda25b43dc5cd90625ee4b0744751799c917",
                "swh:1:dir:5256e856a0a0898966d6ba14feb4388b8b82d302",
            ),
            # complete_snapshot → directory2
            (
                "swh:1:snp:db99fda25b43dc5cd90625ee4b0744751799c917",
                "swh:1:dir:8505808532953da7d2581741f01b29c04b1cb9ab",
            ),
            # complete_snapshot → content
            (
                "swh:1:snp:db99fda25b43dc5cd90625ee4b0744751799c917",
                "swh:1:cnt:d81cc0710eb6cf9efd5b920a8453e1e07157b6cd",
            ),
            # complete_snapshot → revision
            (
                "swh:1:snp:db99fda25b43dc5cd90625ee4b0744751799c917",
                "swh:1:rev:01a7114f36fddd5ef2511b2cadda237a68adbb12",
            ),
            # complete_snapshot → release
            (
                "swh:1:snp:db99fda25b43dc5cd90625ee4b0744751799c917",
                "swh:1:rel:f7f222093a18ec60d781070abec4a630c850b837",
            ),
            # complete_snapshot → empty_snapshot
            (
                "swh:1:snp:db99fda25b43dc5cd90625ee4b0744751799c917",
                "swh:1:snp:1a8893e6a86f444e8be8e7bda6cb34fb1735a00e",
            ),
        },
    )


def test_record_references_adding_origin_visit_statuses(swh_storage, sample_data):
    swh_storage.origin_add([sample_data.origin])

    # The one in sample_data does not reference a snapshot
    origin_visit_status = OriginVisitStatus(
        origin=sample_data.origin.url,
        visit=1,
        date=sample_data.date_visit1,
        type=sample_data.type_visit1,
        status="created",
        snapshot=sample_data.snapshot.id,
    )
    stats = swh_storage.origin_visit_status_add(
        [origin_visit_status, sample_data.origin_visit2_status]
    )

    assert stats["origin_visit_status:add"] > 0
    assert stats["object_reference:add"] == 1

    check_references_recorded(
        swh_storage,
        {
            # origin → snapshot
            (
                "swh:1:ori:33abd4b4c5db79c7387673f71302750fd73e0645",
                "swh:1:snp:9b922e6d8d5b803c1582aabe5525b7b91150788e",
            ),
        },
    )



# Copyright (C) 2020 The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

from collections import Counter, deque
import datetime
from functools import partial
import logging
import traceback
from typing import Callable
from typing import Counter as CounterT
from typing import Deque, Dict, Iterable, List, cast

import yaml

from swh.model.model import BaseModel, Content
from swh.storage import get_storage
from swh.storage.exc import HashCollision

logger = logging.getLogger(__name__)


def now():
    return datetime.datetime.now(datetime.UTC)


class RateQueue:
    def __init__(self, size: int, max_errors: int):
        assert size > max_errors
        self._size = size
        self._max_errors = max_errors
        self._errors: Deque[bool] = deque(maxlen=size)

    def add_ok(self, n_ok: int = 1) -> None:
        self._errors.extend([False] * n_ok)

    def add_error(self, n_error: int = 1) -> None:
        self._errors.extend([True] * n_error)

    def limit_reached(self) -> bool:
        return sum(self._errors) > self._max_errors

    def reset(self):
        # mainly for testing purpose
        self._errors.clear()


class TenaciousProxyStorage:
    """Storage proxy that have a tenacious insertion behavior.

    When an xxx_add method is called, it's first attempted as is against the backend
    storage. If a failure occurs, split the list of inserted objects in pieces until
    erroneous objects have been identified, so all the valid objects are guaranteed to
    be inserted.

    Also provides a error-rate limit feature: if more than n errors occurred during the
    insertion of the last p (window_size) objects, stop accepting any insertion.

    The number of insertion retries for a single object can be specified via
    the 'retries' parameter.

    This proxy is mainly intended to be used in a replayer configuration (aka a
    mirror stack), where insertion errors are mostly unexpected (which explains
    the low default ratio errors/window_size).

    Conversely, it should not be used in a loader configuration, as it may
    drop objects without stopping the loader, which leads to holes in the graph.

    Deployments using this proxy should carefully monitor their logs to check any
    failure is expected (because the failed object is corrupted),
    not because of transient errors or issues with the storage backend.

    Sample configuration use case for tenacious storage:

    .. code-block:: yaml

        storage:
          cls: tenacious
          storage:
            cls: remote
            args: http://storage.internal.staging.swh.network:5002/
          error-rate-limit:
            errors: 10
            window_size: 1000

    """

    tenacious_methods: Dict[str, str] = {
        "content_add": "content",
        "content_add_metadata": "content",
        "skipped_content_add": "skipped_content",
        "directory_add": "directory",
        "revision_add": "revision",
        "extid_add": "extid",
        "release_add": "release",
        "snapshot_add": "snapshot",
        "origin_add": "origin",
    }

    def __init__(
        self,
        storage,
        error_rate_limit: Dict[str, int] | None = None,
        retries: int = 3,
        error_reporter: Callable[[str, bytes], None] | None = None,
    ):
        self.storage = get_storage(**storage)
        if error_rate_limit is None:
            error_rate_limit = {"errors": 10, "window_size": 1000}
        assert "errors" in error_rate_limit
        assert "window_size" in error_rate_limit
        self.rate_queue = RateQueue(
            size=error_rate_limit["window_size"],
            max_errors=error_rate_limit["errors"],
        )
        self._single_object_retries: int = retries
        self.error_reporter = error_reporter

    def __getattr__(self, key):
        if key in self.tenacious_methods:
            return partial(self._tenacious_add, key)
        return getattr(self.storage, key)

    def _tenacious_add(self, func_name, objects: Iterable[BaseModel]) -> Dict[str, int]:
        """Try hard to add as many objects as possible the the backend storage."""
        add_function = getattr(self.storage, func_name)
        object_type = self.tenacious_methods[func_name]

        # list of lists of objects; note this to_add list is consumed from the
        # tail. This list is also deduplicated (while keeping the orders of the
        # elements; using the list(dict.fronkeys(()) trick) to ensure we don't
        # get hit by unicity constraint errors, depending on the actual storage
        # backend we have...
        to_add: List[List[BaseModel]] = [list(dict.fromkeys(objects))]
        n_objs: int = len(to_add[0])

        results: CounterT[str] = Counter()
        retries: int = self._single_object_retries

        while to_add:
            if self.rate_queue.limit_reached():
                logger.error(
                    "Too many insertion errors have been detected; "
                    "disabling insertions"
                )
                raise RuntimeError(
                    "Too many insertion errors have been detected; "
                    "disabling insertions"
                )
            objs = to_add.pop()
            try:
                r = add_function(objs)
                results.update(r)
                self.rate_queue.add_ok(len(objs))
            except HashCollision as exc:
                # In case we have a HashCollision error and the batch on
                # inserted contents only have a few of thems (usually only
                # one), then we can be a bit smarter than the generic logic.
                # This exception gives us the list of failed objects, so use it
                # instead of splitting the batch over and over again
                assert object_type in ("content", "skipped_content")
                to_add.append(
                    [
                        obj
                        for obj in cast(List[Content], objs)
                        if obj.hashes() not in exc.colliding_content_hashes()
                    ]
                )
                dropped_objs = [
                    obj
                    for obj in cast(List[Content], objs)
                    if obj.hashes() in exc.colliding_content_hashes()
                ]
                n_dropped = len(dropped_objs)
                logger.info(
                    "%s: dropping %s %s objects (hash collision)",
                    func_name,
                    n_dropped,
                    object_type,
                )
                for dropped in dropped_objs:
                    logger.info("dropped %s", dropped)
                    if self.error_reporter:
                        key = f"{now().isoformat()}/{object_type}"
                        value = {
                            "exc": traceback.format_exception(exc),
                            "obj": dropped.to_dict(),
                        }
                        self.error_reporter(key, yaml.dump(value).encode())
                results.update({f"{object_type}:add:errors": n_dropped})

            except Exception as exc:
                if len(objs) > 1:
                    logger.info(
                        "%s: failed to insert a batch of %s %s objects, splitting",
                        func_name,
                        len(objs),
                        object_type,
                    )
                    # reinsert objs split in 2 parts at the end of to_add
                    to_add.append(objs[(len(objs) // 2) :])
                    to_add.append(objs[: (len(objs) // 2)])
                    # each time we append a batch in the to_add bag, reset the
                    # one-object-batch retries counter
                    retries = self._single_object_retries
                else:
                    assert len(objs) == 1
                    obj = objs[0]
                    retries -= 1
                    if retries:
                        logger.info(
                            "%s: failed to insert an %s, retrying",
                            func_name,
                            object_type,
                        )
                        # give it another chance
                        to_add.append(objs)
                    else:
                        logger.error(
                            "%s: failed to insert an object, excluding %s (from a batch of %s)",
                            func_name,
                            obj,
                            n_objs,
                        )
                        logger.error(
                            "Exception was: %s",
                            repr(exc),
                            exc_info=not isinstance(exc, HashCollision),
                        )
                        results.update({f"{object_type}:add:errors": 1})
                        if self.error_reporter:
                            key = f"{now().isoformat()}/{object_type}"
                            value = {
                                "obj": obj.to_dict(),
                                "exc": traceback.format_exception(exc),
                            }
                            self.error_reporter(key, yaml.dump(value).encode())
                        self.rate_queue.add_error()
                        # reset the retries counter (needed in case the next
                        # batch is also 1 element only)
                        retries = self._single_object_retries
        return dict(results)

    def reset(self):
        self.rate_queue.reset()



# Copyright (C) 2024-2025 The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information


import pytest

from swh.storage.proxies.blocking.cls import BlockingProxyStorage
from swh.storage.tests.test_in_memory import TestInMemoryStorage as _TestStorage

# This simply test that without any blocked URL defined in the blocking proxy,
# all standard tests are OK


@pytest.fixture
def swh_storage_backend_config():
    yield {
        "cls": "memory",
        "journal_writer": {
            "cls": "memory",
        },
    }


@pytest.fixture
def swh_storage(blocking_db_postgresql, swh_storage_backend):
    storage = BlockingProxyStorage(
        db=blocking_db_postgresql.info.dsn, storage=swh_storage_backend
    )
    try:
        yield storage
    finally:
        storage._blocking_pool.close()


class TestStorage(_TestStorage):
    @pytest.mark.xfail(reason="typing.Protocol instance check is annoying")
    def test_types(self, *args, **kwargs):
        super().test_types(*args, **kwargs)



# Copyright (C) 2020-2024  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

from collections import namedtuple
import datetime
from typing import List

from attr import evolve

from swh.model.hashutil import DEFAULT_ALGORITHMS
from swh.model.model import (
    MetadataAuthority,
    MetadataAuthorityType,
    MetadataFetcher,
    RawExtrinsicMetadata,
)
from swh.model.swhids import ExtendedObjectType, ExtendedSWHID
from swh.storage.cassandra import converters
from swh.storage.cassandra.model import RawExtrinsicMetadataRow

# Test purposes
field_names: List[str] = list(DEFAULT_ALGORITHMS)
Row = namedtuple("Row", field_names)  # type: ignore


def test_row_to_content_hashes():
    for row in [
        Row(
            sha1=b"4\x972t\xcc\xefj\xb4\xdf\xaa\xf8e\x99y/\xa9\xc3\xfeF\x89",
            sha1_git=b"\xd8\x1c\xc0q\x0e\xb6\xcf\x9e\xfd[\x92\n\x84S\xe1\xe0qW\xb6\xcd",  # noqa
            sha256=b"g6P\xf96\xcb;\n/\x93\xce\t\xd8\x1b\xe1\x07H\xb1\xb2\x03\xc1\x9e\x81v\xb4\xee\xfc\x19d\xa0\xcf:",  # noqa
            blake2s256=b"\xd5\xfe\x199We'\xe4,\xfdv\xa9EZ$2\xfe\x7fVf\x95dW}\xd9<B\x80\xe7mf\x1d",  # noqa
        ),
        Row(
            sha1=b"4\x972t\xcc\xefj\xb4\xdf\xaa\xf8e\x99y/\xa9\xc3\xfeF\x89",  # noqa
            sha1_git=b"\xd8\x1c\xc0q\x0e\xb6\xcf\x9e\xfd[\x92\n\x84S\xe1\xe0qW\xb6\xcd",  # noqa
            sha256=b"h6P\xf96\xcb;\n/\x93\xce\t\xd8\x1b\xe1\x07H\xb1\xb2\x03\xc1\x9e\x81v\xb4\xee\xfc\x19d\xa0\xcf:",  # noqa
            blake2s256=b"\xd5\xfe\x199We'\xe4,\xfdv\xa9EZ$2\xfe\x7fVf\x95dW}\xd9<B\x80\xe7mf\x1d",  # noqa
        ),
    ]:
        actual_hashes = converters.row_to_content_hashes(row)

        assert len(actual_hashes) == len(DEFAULT_ALGORITHMS)
        for algo in DEFAULT_ALGORITHMS:
            assert actual_hashes[algo] == getattr(row, algo)


def test_row_to_extrinsic_metadata_conversion():
    """Conversion of row to raw extrinsic metadata should be ok"""
    swhid_template = "swh:1:{type}:43d1b8ba928846e2d2ca2e3b1a5edfb2ae5265f0"
    swhid_target_template = "swh:1:{type}:f25a28146e5089828fe80bf9147184e068b6e09f"

    authority_type = MetadataAuthorityType.REGISTRY
    authority_url = "https://softwareheritage.org/"
    fetcher_name = "swh.loader.package.golang.loader.GolangLoader"
    fetcher_version = "5.17.0"
    origin_url = "https://pkg.go.dev/github.com/aws/aws-sdk-go-v2/service/mgn"
    raw_type = "something"

    def init_raw_row_dict(raw_extrinsic_metadata_d):
        row_raw_d = raw_extrinsic_metadata_d.copy()
        row_raw_d["type"] = raw_type
        row_raw_d.pop("authority", None)
        row_raw_d.pop("fetcher", None)

        row_raw_d["authority_type"] = authority_type.value
        row_raw_d["authority_url"] = authority_url
        row_raw_d["fetcher_name"] = fetcher_name
        row_raw_d["fetcher_version"] = fetcher_version

        # Empty values for "empty" conversion
        for key, empty_value in [
            ("snapshot", ""),
            ("release", ""),
            ("revision", ""),
            ("directory", ""),
            ("visit", 0),
            ("path", b""),
        ]:
            row_raw_d[key] = empty_value

        return row_raw_d

    def init_raw_dict(raw_extrinsic_metadata_d):
        raw_metadata_d = raw_extrinsic_metadata_d.copy()

        for key in ["snapshot", "release", "revision", "directory", "visit", "path"]:
            raw_metadata_d[key] = None

        return raw_metadata_d

    def swhid(obj_type):
        return swhid_template.format(type=ExtendedObjectType[obj_type.upper()].value)

    for key in ["snapshot", "release", "revision", "directory"]:
        if key == "directory":
            target_swhid = swhid_target_template.format(
                type=ExtendedObjectType.CONTENT.value
            )
        else:
            target_swhid = swhid_target_template.format(
                type=ExtendedObjectType.DIRECTORY.value
            )

        # template raw extrinsic metadata we are gonna adapt to ensure conversion is ok
        raw_extrinsic_metadata_d = RawExtrinsicMetadata(
            target=ExtendedSWHID.from_string(target_swhid),
            discovery_date=datetime.datetime(
                2024, 3, 18, 22, 59, 19, tzinfo=datetime.timezone.utc
            ),
            authority=MetadataAuthority(
                type=authority_type, url=authority_url, metadata=None
            ),
            fetcher=MetadataFetcher(
                name=fetcher_name, version=fetcher_version, metadata=None
            ),
            format="original-artifacts-json",
            metadata=b'[{"length": 97190}]',
            origin=origin_url,
        ).to_dict()

        swhid_str = swhid(key)
        raw_metadata_d = init_raw_dict(raw_extrinsic_metadata_d)
        raw_metadata_d[key] = swhid_str
        expected_raw_metadata = RawExtrinsicMetadata.from_dict(raw_metadata_d)

        row_raw_d = init_raw_row_dict(raw_metadata_d)
        row_raw_d[key] = swhid_str
        row_raw = RawExtrinsicMetadataRow.from_dict(row_raw_d)

        actual_raw = converters.row_to_raw_extrinsic_metadata(row_raw)

        # As we crafted an extrinsic metadata for test conversion purpose, we do not
        # care for the id object comparison. We just ensure the conversion is ok.
        assert actual_raw == evolve(expected_raw_metadata, id=actual_raw.id)



# Copyright (C) 2019-2025  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import atexit
import datetime
from functools import partial
import os
import re
import resource
import shutil
import signal
import socket
import subprocess
import sys
import time
from typing import List, Optional
import uuid

from cassandra.auth import PlainTextAuthProvider
from cassandra.cluster import Cluster
import pytest
from pytest_postgresql import factories
from pytest_shared_session_scope import (
    CleanupToken,
    SetupToken,
    shared_session_scope_json,
)

from swh.core.db.db_utils import initialize_database_for_module
from swh.storage import get_storage
from swh.storage.postgresql.storage import Storage as StorageDatastore
from swh.storage.tests.storage_data import StorageData

os.environ["LC_ALL"] = "C.UTF-8"


_CASSANDRA_CONFIG_TEMPLATE = """
data_file_directories:
    - {data_dir}/data
commitlog_directory: {data_dir}/commitlog
hints_directory: {data_dir}/hints
saved_caches_directory: {data_dir}/saved_caches

commitlog_sync: periodic
commitlog_sync_period_in_ms: 1000000
partitioner: org.apache.cassandra.dht.Murmur3Partitioner
endpoint_snitch: SimpleSnitch
seed_provider:
    - class_name: org.apache.cassandra.locator.SimpleSeedProvider
      parameters:
          - seeds: "127.0.0.1"

storage_port: {storage_port}
native_transport_port: {native_transport_port}
start_native_transport: true
listen_address: 127.0.0.1

enable_user_defined_functions: true

# speed-up by disabling period saving to disk
key_cache_save_period: 0
row_cache_save_period: 0
trickle_fsync: false
commitlog_sync_period_in_ms: 100000
authenticator: PasswordAuthenticator
"""

_SCYLLA_EXTRA_CONFIG_TEMPLATE = """
experimental_features:
    - udf
view_hints_directory: {data_dir}/view_hints
prometheus_port: 0  # disable prometheus server
start_rpc: false  # disable thrift server
api_port: {api_port}
"""


def _free_port():
    sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    sock.bind(("127.0.0.1", 0))
    port = sock.getsockname()[1]
    sock.close()
    return port


def _java_major_version() -> Optional[int]:
    java_bins: List[Optional[str]] = []
    if "JAVA_HOME" in os.environ:
        java_bins.append(f'{os.environ["JAVA_HOME"].rstrip("/")}/bin/java')
    java_bins.append(shutil.which("java"))
    for java_bin in java_bins:
        if java_bin and os.access(java_bin, os.X_OK):
            break
    else:
        raise EnvironmentError(
            "Unable to find java executable. Check JAVA_HOME and PATH environment variables."
        )
    version = subprocess.check_output(
        [java_bin, "-version"],
        stderr=subprocess.STDOUT,
        text=True,
    )
    match = re.search(r'"(\d+\.\d+).*"', version)
    if match:
        return int(float(match.groups()[0]))
    return None


def _wait_for_peer(addr, port):
    wait_until = time.time() + 60
    while time.time() < wait_until:
        try:
            sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
            sock.connect((addr, port))
        except ConnectionRefusedError:
            time.sleep(0.1)
        else:
            sock.close()
            return True
    return False


def _kill_cassandra_cluster(pgrp):
    os.killpg(pgrp, signal.SIGKILL)


@pytest.fixture(scope="session")
def cassandra_auth_provider_config():
    return {
        "cls": "cassandra.auth.PlainTextAuthProvider",
        "username": "cassandra",
        "password": "cassandra",
    }


@pytest.fixture(scope="session")
def session_uuid():
    return os.environ.get("PYTEST_XDIST_TESTRUNUID", str(uuid.uuid4))


@shared_session_scope_json(
    params=[
        pytest.param("", marks=pytest.mark.cassandra),
    ]
)
def swh_storage_cassandra_cluster(tmpdir_factory, tmp_path_factory, session_uuid):
    cassandra_conf = tmpdir_factory.mktemp("cassandra_conf")
    cassandra_data = tmpdir_factory.mktemp("cassandra_data")
    cassandra_log = tmpdir_factory.mktemp("cassandra_log")
    root_tmp_dir = tmp_path_factory.getbasetemp().parent
    proc = None

    data = yield

    if data == SetupToken.FIRST:
        # first pytest-xdist worker to execute that session scope fixture
        # spawns the cassandra process

        native_transport_port = _free_port()
        storage_port = _free_port()
        jmx_port = _free_port()
        api_port = _free_port()

        use_scylla = bool(os.environ.get("SWH_USE_SCYLLADB", ""))

        cassandra_bin = os.environ.get(
            "SWH_CASSANDRA_BIN",
            "/usr/bin/scylla" if use_scylla else "/usr/sbin/cassandra",
        )

        if use_scylla:
            os.makedirs(cassandra_conf.join("conf"))
            config_path = cassandra_conf.join("conf/scylla.yaml")
            config_template = _CASSANDRA_CONFIG_TEMPLATE + _SCYLLA_EXTRA_CONFIG_TEMPLATE
        else:
            config_path = cassandra_conf.join("cassandra.yaml")
            config_template = _CASSANDRA_CONFIG_TEMPLATE

        with open(str(config_path), "w") as fd:
            fd.write(
                config_template.format(
                    data_dir=str(cassandra_data),
                    storage_port=storage_port,
                    native_transport_port=native_transport_port,
                    api_port=api_port,
                )
            )

        if os.environ.get("SWH_CASSANDRA_LOG"):
            stdout = stderr = None
        else:
            stdout = stderr = subprocess.DEVNULL

        env = {
            "MAX_HEAP_SIZE": "300M",
            "HEAP_NEWSIZE": "50M",
            "JVM_OPTS": f"-Xlog:gc=error:file={cassandra_log}/gc.log",
            "CASSANDRA_LOG_DIR": cassandra_log,
        }

        java_major_version = _java_major_version()
        if java_major_version and java_major_version >= 17:
            env["JVM_OPTS"] += " -Djava.security.manager=allow"

        if "JAVA_HOME" in os.environ:
            env["JAVA_HOME"] = os.environ["JAVA_HOME"]

        if use_scylla:
            env = {
                **env,
                "SCYLLA_HOME": cassandra_conf,
            }
            # prevent "NOFILE rlimit too low (recommended setting 200000,
            # minimum setting 10000; refusing to start."
            resource.setrlimit(resource.RLIMIT_NOFILE, (200000, 200000))

            proc = subprocess.Popen(
                [
                    cassandra_bin,
                    "--developer-mode=1",
                ],
                start_new_session=True,
                env=env,
                stdout=stdout,
                stderr=stderr,
            )
        else:
            proc = subprocess.Popen(
                [
                    cassandra_bin,
                    "-Dcassandra.config=file://%s/cassandra.yaml" % cassandra_conf,
                    "-Dcassandra.logdir=%s" % cassandra_log,
                    "-Dcassandra.jmx.local.port=%d" % jmx_port,
                    "-Dcassandra-foreground=yes",
                ],
                start_new_session=True,
                env=env,
                stdout=stdout,
                stderr=stderr,
            )

        listening = _wait_for_peer("127.0.0.1", native_transport_port)

        if listening:
            # Wait for initialization
            auth_provider = PlainTextAuthProvider(
                username="cassandra", password="cassandra"
            )
            with Cluster(
                ["127.0.0.1"],
                port=native_transport_port,
                auth_provider=auth_provider,
                connect_timeout=60,
                control_connection_timeout=60,
            ) as cluster:

                session = None
                retry = 0
                while (not session) and retry < 10:
                    try:
                        session = cluster.connect()
                    except Exception:
                        time.sleep(1)
                        retry += 1

        # ensure to kill cassandra cluster when test runner process exits
        atexit.register(_kill_cassandra_cluster, os.getpgid(proc.pid))

        data = (["127.0.0.1"], native_transport_port)

    token: CleanupToken = yield data

    if token == CleanupToken.LAST:
        # last pytest-xdist worker tearing down that session scope fixture
        # informs the worker that spawned cassandra it can be shutdowned
        (root_tmp_dir / session_uuid).touch()

    if proc is not None:
        if not listening:
            if proc.poll() is None:
                raise Exception("cassandra process unexpectedly not listening.")
            else:
                raise Exception("cassandra process unexpectedly stopped.")

        while not (root_tmp_dir / session_uuid).exists():
            # wait until all pytest-xdist workers executed their test suites
            time.sleep(1)

        if not listening or os.environ.get("SWH_CASSANDRA_LOG"):
            debug_log_path = str(cassandra_log.join("debug.log"))
            if os.path.exists(debug_log_path):
                with open(debug_log_path) as fd:
                    print(fd.read(), file=sys.stderr)


@pytest.fixture(scope="session")
def swh_storage_cassandra_keyspace(
    swh_storage_cassandra_cluster, cassandra_auth_provider_config
):
    from swh.storage.cassandra import create_keyspace
    from swh.storage.cassandra.cql import CqlRunner

    (hosts, port) = swh_storage_cassandra_cluster
    keyspace = "test" + os.urandom(10).hex()

    cql_runner = CqlRunner(
        hosts,
        keyspace,
        port,
        auth_provider=cassandra_auth_provider_config,
        consistency_level="ONE",
        register_user_types=False,
    )

    create_keyspace(cql_runner)

    cql_runner._cluster.shutdown()

    return keyspace


@pytest.fixture
def swh_storage_cassandra_backend_config(
    swh_storage_cassandra_cluster,
    swh_storage_cassandra_keyspace,
    cassandra_auth_provider_config,
):

    (hosts, port) = swh_storage_cassandra_cluster

    keyspace = swh_storage_cassandra_keyspace

    storage_config = dict(
        cls="cassandra",
        hosts=hosts,
        port=port,
        keyspace=keyspace,
        journal_writer={"cls": "memory"},
        objstorage={"cls": "memory"},
        auth_provider=cassandra_auth_provider_config,
    )

    yield storage_config


swh_storage_postgresql_proc = factories.postgresql_proc(
    load=[
        partial(
            initialize_database_for_module,
            modname="storage",
            version=StorageDatastore.current_version,
        ),
    ],
)


swh_storage_postgresql = factories.postgresql(
    "swh_storage_postgresql_proc",
)


@pytest.fixture
def swh_storage_postgresql_backend_config(swh_storage_postgresql):
    """Basic pg storage configuration with no journal collaborator
    (to avoid pulling optional dependency on clients of this fixture)

    """
    yield {
        "cls": "postgresql",
        "db": swh_storage_postgresql.info.dsn,
        "objstorage": {"cls": "memory"},
        "check_config": {"check_write": True},
        "max_pool_conns": 100,
    }


@pytest.fixture
def swh_storage_backend_config(swh_storage_postgresql_backend_config):
    """Configuration to use for :func:`swh_storage_backend`.

    Defaults to :func:`swh_storage_postgresql_backend_config`.

    """
    return swh_storage_postgresql_backend_config


@pytest.fixture
def swh_storage_backend(swh_storage_backend_config):
    """
    By default, this fixture aliases ``swh_storage``. However, when ``swh_storage``
    is overridden to be a proxy storage, this fixture returns the storage instance
    behind all proxies.

    This is useful to introspect the state of backends from proxy tests"""
    storage = get_storage(**swh_storage_backend_config)

    backend = storage

    # handle storage pipeline for backward-compatibility as
    # object_references_create_partition is only available on
    # real storage backend, not on proxies.
    while hasattr(backend, "storage"):
        backend = backend.storage

    backend.object_references_create_partition(
        *datetime.date.today().isocalendar()[0:2]
    )

    cassandra_backend = hasattr(backend, "_cql_runner") and hasattr(
        backend._cql_runner, "_cluster"
    )

    if cassandra_backend:
        from swh.storage.cassandra.cql import mark_all_migrations_completed

        mark_all_migrations_completed(backend._cql_runner)

    yield storage

    if cassandra_backend:
        from swh.storage.cassandra.schema import TABLES

        keyspace = backend._keyspace
        for table in TABLES:
            table_rows = backend._cql_runner.execute_with_retries(
                f"SELECT * from {keyspace}.{table} LIMIT 1", args=[]
            )
            if table_rows.one() is not None:
                backend._cql_runner.execute_with_retries(
                    f"TRUNCATE TABLE {keyspace}.{table}", args=[]
                )
        backend._cql_runner._cluster.shutdown()
    if hasattr(backend, "_pool"):
        backend._pool.close()


@pytest.fixture
def swh_storage(swh_storage_backend):
    return swh_storage_backend


@pytest.fixture
def sample_data() -> StorageData:
    """Pre-defined sample storage object data to manipulate

    Returns:
        StorageData whose attribute keys are data model objects. Either multiple
        objects: contents, directories, revisions, releases, ... or simple ones:
        content, directory, revision, release, ...

    """
    return StorageData()



from typing import Iterable, List

from swh.model import discovery, model
from swh.model.model import Sha1Git
from swh.storage.interface import StorageInterface


class DiscoveryStorageConnection(discovery.ArchiveDiscoveryInterface):
    """Use the storage APIs to query the archive"""

    def __init__(
        self,
        contents: List[model.Content],
        skipped_contents: List[model.SkippedContent],
        directories: List[model.Directory],
        swh_storage: StorageInterface,
    ) -> None:
        self.contents = contents
        self.skipped_contents = skipped_contents
        self.directories = directories
        self.storage = swh_storage

    def content_missing(self, contents: List[Sha1Git]) -> Iterable[Sha1Git]:
        """List content missing from the archive by sha1"""
        return self.storage.content_missing_per_sha1_git(contents)

    def skipped_content_missing(
        self, skipped_contents: List[Sha1Git]
    ) -> Iterable[Sha1Git]:
        """List skipped content missing from the archive by sha1"""
        contents = [
            {"sha1_git": s, "sha1": None, "sha256": None, "blake2s256": None}
            for s in skipped_contents
        ]
        return (d["sha1_git"] for d in self.storage.skipped_content_missing(contents))

    def directory_missing(self, directories: List[Sha1Git]) -> Iterable[Sha1Git]:
        """List directories missing from the archive by sha1"""
        return self.storage.directory_missing(directories)



# Copyright (C) 2019-2025  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information


CREATE_TABLES_QUERIES = {
    "microtimestamp": """
CREATE TYPE IF NOT EXISTS microtimestamp (
    seconds             bigint,
    microseconds        int,
);""",
    "microtimestamp_with_timezone": """
CREATE TYPE IF NOT EXISTS microtimestamp_with_timezone (
    timestamp           frozen<microtimestamp>,
    offset_bytes        blob,
);""",
    "person": """
CREATE TYPE IF NOT EXISTS person (
    fullname    blob,
    name        blob,
    email       blob
);""",
    "migration": """
CREATE TABLE IF NOT EXISTS migration (
    id                  ascii,
    dependencies        frozen<set<ascii>>,
    min_read_version    ascii,
    status              ascii,
    PRIMARY KEY ((id))
) WITH
    comment = 'Set of known database migrations'
    {table_options};""",
    "content": """
CREATE TABLE IF NOT EXISTS content (
    sha1          blob,
    sha1_git      blob,
    sha256        blob,
    blake2s256    blob,
    length        bigint,
    ctime         timestamp,
    status        ascii,
    PRIMARY KEY ((sha256), sha1, sha1_git, blake2s256)
) WITH
    comment = 'The content of a file, without a file name'
    {table_options};""",
    "skipped_content": """
CREATE TABLE IF NOT EXISTS skipped_content (
    sha1          blob,
    sha1_git      blob,
    sha256        blob,
    blake2s256    blob,
    length        bigint,
    ctime         timestamp,
    status        ascii,
    reason        text,
    origin        text,
    PRIMARY KEY ((sha1, sha1_git, sha256, blake2s256))
) WITH
    comment = 'A content that could not be archived by SWH for a given reason'
    {table_options};""",
    "revision": """
CREATE TABLE IF NOT EXISTS revision (
    id                              blob,
    date                            microtimestamp_with_timezone,
    committer_date                  microtimestamp_with_timezone,
    type                            ascii,
    directory                       blob,
    message                         blob,
    author                          person,
    committer                       person,
    synthetic                       boolean,
    metadata                        text,
    extra_headers                   frozen<list <list<blob>> >,
    raw_manifest                    blob,
    PRIMARY KEY ((id))
) WITH
    comment = 'Intermediary state of development of a project'
    {table_options};""",
    "revision_parent": """
CREATE TABLE IF NOT EXISTS revision_parent (
    id                     blob,
    parent_rank                     int,
    parent_id                       blob,
    PRIMARY KEY ((id), parent_rank)
) WITH
    comment = 'Ordered list of parents of a revision'
    {table_options};""",
    "release": """
CREATE TABLE IF NOT EXISTS release (
    id                              blob,
    target_type                     ascii,
    target                          blob,
    date                            microtimestamp_with_timezone,
    name                            blob,
    message                         blob,
    author                          person,
    synthetic                       boolean,
    raw_manifest                    blob,
    PRIMARY KEY ((id))
) WITH
    comment = 'Published versions of a software project'
    {table_options};""",
    "directory": """
CREATE TABLE IF NOT EXISTS directory (
    id              blob,
    raw_manifest                    blob,
    PRIMARY KEY ((id))
) WITH
    comment = 'Set of named references to other directories, files, and revisions'
    {table_options};""",
    "directory_entry": """
CREATE TABLE IF NOT EXISTS directory_entry (
    directory_id    blob,
    name            blob,
    target          blob,
    perms           int,
    type            ascii,
    PRIMARY KEY ((directory_id), name)
) WITH
    comment = 'Named reference from a directory to another entry (subdirectory, file, or revision)'
    {table_options};""",  # noqa: B950
    "snapshot": """
CREATE TABLE IF NOT EXISTS snapshot (
    id              blob,
    PRIMARY KEY ((id))
) WITH
    comment = 'State of an origin at a given time'
    {table_options};""",
    "snapshot_branch": """
CREATE TABLE IF NOT EXISTS snapshot_branch (
    snapshot_id     blob,
    name            blob,
    target_type     ascii,
    target          blob,
    PRIMARY KEY ((snapshot_id), name)
) WITH
    comment = 'Named referenced from a snapshot to an other object'
    {table_options};""",
    "origin_visit": """
CREATE TABLE IF NOT EXISTS origin_visit (
    origin          text,
    visit           bigint,
    date            timestamp,
    type            text,
    PRIMARY KEY ((origin), visit)
) WITH
    comment = 'Times where SWH looked at an origin for new content'
    {table_options};""",
    "origin_visit_status": """
CREATE TABLE IF NOT EXISTS origin_visit_status (
    origin          text,
    visit           bigint,
    date            timestamp,
    type            text,
    status          ascii,
    metadata        text,
    snapshot        blob,
    PRIMARY KEY ((origin), visit, date)
) WITH
    CLUSTERING ORDER BY (visit DESC, date DESC)
    AND comment = 'Updates to an origin visit'
    {table_options};""",  # 'WITH CLUSTERING ORDER BY' is optional with Cassandra 4,
    # but ScyllaDB needs it
    "origin": """
CREATE TABLE IF NOT EXISTS origin (
    sha1            blob,
    url             text,
    next_visit_id   int,
    PRIMARY KEY ((sha1))
) WITH
    comment = 'Software repositories'
    {table_options};""",
    "metadata_authority": """
CREATE TABLE IF NOT EXISTS metadata_authority (
    url             text,
    type            ascii,
    PRIMARY KEY ((url), type)
) WITH
    comment = 'External source of objects present in the "raw_extrinsic_metadata" table'
    {table_options};""",
    "metadata_fetcher": """
CREATE TABLE IF NOT EXISTS metadata_fetcher (
    name            ascii,
    version         ascii,
    PRIMARY KEY ((name), version)
) WITH
    comment = 'Software running on SWH that fetches objects present in the "raw_extrinsic_metadata" table'
    {table_options};""",  # noqa: B950
    # Cassandra handles null values for row properties as removals (tombstones), which
    # are never cleaned up as the values were never set. To avoid this issue, we instead
    # store a known invalid value of the proper type as a placeholder for null
    # properties: for strings and bytes: the empty value; for visit ids (integers): 0.
    # See comments for visit, snapshot, release, revision, path and directory fields
    "raw_extrinsic_metadata": """
CREATE TABLE IF NOT EXISTS raw_extrinsic_metadata (
    id              blob,
    type            text,
    target          text,
    authority_type  text,
    authority_url   text,
    discovery_date  timestamp,
    fetcher_name    ascii,
    fetcher_version ascii,
    format          ascii,
    metadata        blob,
    origin          text,
    visit           bigint,
    snapshot        text,
    release         text,
    revision        text,
    path            blob,
    directory       text,
    PRIMARY KEY ((target), authority_type, authority_url, discovery_date, id)
) WITH
    comment = 'Blobs of metadata obtained from external sources, that do not fit in the main data model'
    {table_options};""",  # noqa: B950
    "raw_extrinsic_metadata_by_id": """
CREATE TABLE IF NOT EXISTS raw_extrinsic_metadata_by_id (
    id              blob,
    target          text,
    authority_type  text,
    authority_url   text,
    PRIMARY KEY ((id))
) WITH
    comment = 'Secondary table to access "raw_extrinsic_metadata_by_id" from the swh:1:emd: pseudo-SWHID'
    {table_options};""",  # noqa: B950
    "extid": """
CREATE TABLE IF NOT EXISTS extid (
    extid_type      ascii,
    extid           blob,
    extid_version   smallint,
    target_type     ascii,
    target          blob,
    PRIMARY KEY ((extid_type, extid), extid_version, target_type, target)
) WITH
    comment = 'External identifiers for objects with SWHIDs'
    {table_options};""",
    "extid_by_target": """
CREATE TABLE IF NOT EXISTS extid_by_target (
    target_type     ascii,
    target          blob,
    target_token    bigint,
    PRIMARY KEY ((target_type, target), target_token)
) WITH
    comment = ''
    {table_options};""",
    "object_references": """
CREATE TABLE IF NOT EXISTS object_references (
    target_type     ascii,
    target          blob,
    source_type     ascii,
    source          blob,
    PRIMARY KEY ((target_type, target), source_type, source)
) WITH
    comment = 'Set of reverse-references, ie. set of sources pointing to a target. Not written to anymore, but still supported for reads'
    {table_options};""",  # noqa: B950
    "object_references_table": """
CREATE TABLE IF NOT EXISTS object_references_table (
    pk              int, -- always zero, puts everything in the same Cassandra partition
    name            ascii,
    year            int, -- ISO year
    week            int, -- ISO week
    start           date,
    end             date,
    PRIMARY KEY ((pk), name)
) WITH
    comment = 'Stores the list of object_references_* tables; which are dynamically created and dropped'
    {table_options};""",  # noqa: B950
}

CONTENT_INDEX_TEMPLATE = """
CREATE TABLE IF NOT EXISTS content_by_{main_algo} (
    {main_algo}   blob,
    target_token  bigint, -- value of token(pk) on the "primary" table
    PRIMARY KEY (({main_algo}), target_token)
) WITH
    comment = 'Secondary table, used for looking up "content" from a single hash'
    {{table_options}};"""

SKIPPED_CONTENT_INDEX_TEMPLATE = """
CREATE TABLE IF NOT EXISTS skipped_content_by_{main_algo} (
    {main_algo}   blob,
    target_token  bigint, -- value of token(pk) on the "primary" table
    PRIMARY KEY (({main_algo}), target_token)
) WITH
    comment = 'Secondary table, used for looking up "content" from a single hash'
    {{table_options}};"""

OBJECT_REFERENCES_TEMPLATE = """
CREATE TABLE IF NOT EXISTS {keyspace}.{name} (
    target_type     ascii,
    target          blob,
    source_type     ascii,
    source          blob,
    PRIMARY KEY ((target_type, target), source_type, source)
) WITH
    comment = 'Time-based shard of the set of reverse-references (ie. set of sources pointing to a target)'
    {table_options};"""  # noqa: B950

TABLES = [
    "migration",
    "skipped_content",
    "content",
    "revision",
    "revision_parent",
    "release",
    "directory",
    "directory_entry",
    "snapshot",
    "snapshot_branch",
    "origin_visit",
    "origin",
    "raw_extrinsic_metadata",
    "raw_extrinsic_metadata_by_id",
    "origin_visit_status",
    "metadata_authority",
    "metadata_fetcher",
    "extid",
    "extid_by_target",
    "object_references",
    "object_references_table",
]


HASH_ALGORITHMS = ["sha1", "sha1_git", "sha256", "blake2s256"]

for main_algo in HASH_ALGORITHMS:
    TABLES.append(f"content_by_{main_algo}")
    CREATE_TABLES_QUERIES[f"content_by_{main_algo}"] = CONTENT_INDEX_TEMPLATE.format(
        main_algo=main_algo,
        other_algos=", ".join([algo for algo in HASH_ALGORITHMS if algo != main_algo]),
    )

    TABLES.append(f"skipped_content_by_{main_algo}")
    CREATE_TABLES_QUERIES[f"skipped_content_by_{main_algo}"] = (
        SKIPPED_CONTENT_INDEX_TEMPLATE.format(
            main_algo=main_algo,
            other_algos=", ".join(
                [algo for algo in HASH_ALGORITHMS if algo != main_algo]
            ),
        )
    )



# Copyright (C) 2019-2024  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

from datetime import datetime, timezone
from typing import Callable, Dict, Optional, Tuple, TypeVar

from swh.model.hashutil import DEFAULT_ALGORITHMS, hash_to_bytes, hash_to_hex


def now() -> datetime:
    return datetime.now(tz=timezone.utc)


T1 = TypeVar("T1")
T2 = TypeVar("T2")


def map_optional(f: Callable[[T1], T2], x: Optional[T1]) -> Optional[T2]:
    """Execute function f on x if x is not None and return the result. Return None
    otherwise.

    Args:
        f: Function to be called on non-None value
        x: The value to apply f onto (or None)

    Returns:
        f(x) if x is not None else None.

    """
    return f(x) if x is not None else None


def _is_power_of_two(n: int) -> bool:
    return n > 0 and n & (n - 1) == 0


def get_partition_bounds_bytes(
    i: int, n: int, nb_bytes: int
) -> Tuple[bytes, Optional[bytes]]:
    r"""Splits the range [0; 2^(nb_bytes*8)) into n same-length intervals,
    and returns the boundaries of this interval (both inclusive); or None
    as upper bound, if this is the last partition

    n must be a power of 2.

    >>> get_partition_bounds_bytes(0, 16, 2) == (b'\x00\x00', b'\x10\x00')
    True
    >>> get_partition_bounds_bytes(1, 16, 2) == (b'\x10\x00', b'\x20\x00')
    True
    >>> get_partition_bounds_bytes(14, 16, 2) == (b'\xe0\x00', b'\xf0\x00')
    True
    >>> get_partition_bounds_bytes(15, 16, 2) == (b'\xf0\x00', None)
    True
    """
    if not _is_power_of_two(n):
        raise ValueError("number of partitions must be a power of two")
    if not 0 <= i < n:
        raise ValueError(
            "partition index must be between 0 and the number of partitions."
        )

    space_size = 1 << (nb_bytes * 8)
    partition_size = space_size // n

    start = (partition_size * i).to_bytes(nb_bytes, "big")
    end = None if i == n - 1 else (partition_size * (i + 1)).to_bytes(nb_bytes, "big")
    return (start, end)


def content_hex_hashes(content: Dict[str, bytes]) -> Dict[str, str]:
    """Convert bytes hashes into hex hashes."""
    return {algo: hash_to_hex(content[algo]) for algo in DEFAULT_ALGORITHMS}


def content_bytes_hashes(content: Dict[str, str]) -> Dict[str, bytes]:
    """Convert bytes hashes into hex hashes."""
    return {algo: hash_to_bytes(content[algo]) for algo in DEFAULT_ALGORITHMS}


def remove_keys(d: Dict[T1, T2], keys: Tuple[T1, ...]) -> Dict[T1, T2]:
    """Returns a copy of ``d`` minus the given keys."""
    return {k: v for (k, v) in d.items() if k not in keys}


def round_to_milliseconds(date):
    """Round datetime to milliseconds before insertion, so equality doesn't fail after a
    round-trip through a DB (eg. Cassandra)

    """
    return date.replace(microsecond=(date.microsecond // 1000) * 1000)



# Copyright (C) 2015-2025  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

from functools import partial
import logging
import os
from typing import Any, Dict, Optional

from swh.core import config
from swh.core.api import RPCServerApp
from swh.core.api import encode_data_server as encode_data
from swh.core.api import error_handler
from swh.storage import get_storage as get_swhstorage

from ..exc import NonRetryableException, QueryTimeout
from ..interface import StorageInterface
from ..metrics import send_metric, timed
from .serializers import DECODERS, ENCODERS


def get_storage():
    global storage
    if not storage:
        storage = get_swhstorage(**app.config["storage"])

    return storage


class StorageServerApp(RPCServerApp):
    extra_type_decoders = DECODERS
    extra_type_encoders = ENCODERS

    method_decorators = [timed]

    def _process_metrics(self, metrics, endpoint):
        for metric, count in metrics.items():
            send_metric(metric=metric, count=count, method_name=endpoint)

    def post_content_add(self, ret, kw):
        self._process_metrics(ret, "content_add")

    def post_content_add_metadata(self, ret, kw):
        self._process_metrics(ret, "content_add_metadata")

    def post_skipped_content_add(self, ret, kw):
        self._process_metrics(ret, "skipped_content_add")

    def post_directory_add(self, ret, kw):
        self._process_metrics(ret, "directory_add")

    def post_revision_add(self, ret, kw):
        self._process_metrics(ret, "revision_add")

    def post_release_add(self, ret, kw):
        self._process_metrics(ret, "release_add")

    def post_snapshot_add(self, ret, kw):
        self._process_metrics(ret, "snapshot_add")

    def post_origin_visit_status_add(self, ret, kw):
        self._process_metrics(ret, "origin_visit_status_add")

    def post_origin_add(self, ret, kw):
        self._process_metrics(ret, "origin_add")

    def post_raw_extrinsic_metadata_add(self, ret, kw):
        self._process_metrics(ret, "raw_extrinsic_metadata_add")

    def post_metadata_fetcher_add(self, ret, kw):
        self._process_metrics(ret, "metadata_fetcher_add")

    def post_metadata_authority_add(self, ret, kw):
        self._process_metrics(ret, "metadata_authority_add")

    def post_extid_add(self, ret, kw):
        self._process_metrics(ret, "extid_add")

    def post_origin_visit_add(self, ret, kw):
        nb_visits = len(ret)
        send_metric(
            "origin_visit:add",
            count=nb_visits,
            # method_name should be "origin_visit_add", but changing it now would break
            # existing metrics
            method_name="origin_visit",
        )


app = StorageServerApp(
    __name__, backend_class=StorageInterface, backend_factory=get_storage
)
storage = None


@app.errorhandler(QueryTimeout)
def query_timeout_handler(exception):
    """Return 503 instead of 500, telling clients this is a transient
    error and they should retry."""
    return error_handler(exception, encode_data, status_code=503)


@app.errorhandler(NonRetryableException)
def non_retryable_error_handler(exception):
    """Send all non-retryable errors with a 400 status code so the client can
    re-raise them."""
    return error_handler(
        exception, partial(encode_data, extra_type_encoders=ENCODERS), status_code=400
    )


@app.errorhandler(Exception)
def default_error_handler(exception):
    return error_handler(exception, encode_data)


@app.route("/")
@timed
def index():
    return """<html>
<head><title>Software Heritage storage server</title></head>
<body>
<p>You have reached the
<a href="https://www.softwareheritage.org/">Software Heritage</a>
storage server.<br />
See its
<a href="https://docs.softwareheritage.org/devel/swh-storage/">documentation
and API</a> for more information</p>
</body>
</html>"""


@app.route("/stat/counters", methods=["GET"])
@timed
def stat_counters():
    return encode_data(get_storage().stat_counters())


@app.route("/stat/refresh", methods=["GET"])
@timed
def refresh_stat_counters():
    return encode_data(get_storage().refresh_stat_counters())


api_cfg = None


def load_and_check_config(config_path: Optional[str]) -> Dict[str, Any]:
    """Check the minimal configuration is set to run the api or raise an
       error explanation.

    Args:
        config_path: Path to the configuration file to load

    Raises:
        Error if the setup is not as expected

    Returns:
        configuration as a dict

    """
    if not config_path:
        raise EnvironmentError("Configuration file must be defined")

    if not os.path.exists(config_path):
        raise FileNotFoundError(f"Configuration file {config_path} does not exist")

    cfg = config.read(config_path)
    if "storage" not in cfg:
        raise KeyError("Missing 'storage' configuration")

    return cfg


def make_app_from_configfile() -> StorageServerApp:
    """Run the WSGI app from the webserver, loading the configuration from
    a configuration file.

    SWH_CONFIG_FILENAME environment variable defines the
    configuration path to load.

    """
    global api_cfg
    if not api_cfg:
        config_path = os.environ.get("SWH_CONFIG_FILENAME")
        api_cfg = load_and_check_config(config_path)
        app.config.update(api_cfg)
    handler = logging.StreamHandler()
    app.logger.addHandler(handler)
    return app


if __name__ == "__main__":
    print("Deprecated. Use swh-storage")



# Copyright (C) 2015-2025  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

from contextlib import contextmanager
import queue
import threading
import time
from typing import List
from unittest.mock import Mock

import attr
import pytest

from swh.model import from_disk
from swh.model.model import Directory, Person
from swh.storage.tests.storage_tests import TestStorage as _TestStorage
from swh.storage.tests.storage_tests import TestStorageDeletion as _TestStorageDeletion
from swh.storage.tests.storage_tests import TestStorageGeneratedData  # noqa
from swh.storage.utils import now


@contextmanager
def db_transaction(storage):
    with storage.db() as db:
        with db.transaction() as cur:
            yield db, cur


def test_pgstorage_flavor(swh_storage):
    assert swh_storage.get_flavor() == "default"


class TestStorage(_TestStorage):
    @pytest.mark.skip(
        "Directory pagination is not implemented in the postgresql backend yet."
    )
    def test_directory_get_entries_pagination(self):
        pass


class TestStorageDeletion(_TestStorageDeletion):
    def _affected_tables(self) -> List[str]:
        return [
            "origin",
            "origin_visit",
            "origin_visit_status",
            "snapshot",
            # `snapshot_branch` is left out, we leave stale data there by
            # design.
            "snapshot_branches",
            "release",
            "revision",
            "revision_history",
            "directory",
            # `directory_entry_*` are left out on purpose.
            # We leave stale data there by design.
            "skipped_content",
            "content",
            # `metadata_authority` and `metadata_fetcher` are left out
            # for the time being. Tracked as:
            # https://gitlab.softwareheritage.org/swh/devel/swh-alter/-/issues/21
            "raw_extrinsic_metadata",
        ]

    def _count_from_table(self, swh_storage_backend, table: str) -> int:
        with db_transaction(swh_storage_backend) as (_, cur):
            cur.execute(f"SELECT COUNT(*) FROM {table}")
            return cur.fetchone()[0]


class TestPostgresqlStorage:
    """PostgreSQL-specific tests"""

    # This test is only relevant on the local storage, with an actual
    # objstorage raising an exception
    def test_content_add_objstorage_exception(self, swh_storage, sample_data):
        content = sample_data.content

        swh_storage.objstorage.content_add = Mock(
            side_effect=Exception("mocked broken objstorage")
        )

        with pytest.raises(Exception, match="mocked broken"):
            swh_storage.content_add([content])

        missing = list(swh_storage.content_missing([content.hashes()]))
        assert missing == [content.sha1]


class TestStorageRaceConditions:
    @pytest.mark.xfail
    def test_content_add_race(self, swh_storage, sample_data):
        content = attr.evolve(sample_data.content, ctime=now())

        results = queue.Queue()

        def thread():
            try:
                with db_transaction(swh_storage) as (db, cur):
                    ret = swh_storage._content_add_metadata(db, cur, [content])
                results.put((threading.get_ident(), "data", ret))
            except Exception as e:
                results.put((threading.get_ident(), "exc", e))

        t1 = threading.Thread(target=thread)
        t2 = threading.Thread(target=thread)
        t1.start()
        # this avoids the race condition
        # import time
        # time.sleep(1)
        t2.start()
        t1.join()
        t2.join()

        r1 = results.get(block=False)
        r2 = results.get(block=False)

        with pytest.raises(queue.Empty):
            results.get(block=False)
        assert r1[0] != r2[0]
        assert r1[1] == "data", "Got exception %r in Thread%s" % (r1[2], r1[0])
        assert r2[1] == "data", "Got exception %r in Thread%s" % (r2[2], r2[0])

    def test_directory_add_race(self, swh_storage, sample_data):
        # generate a bunch of similar/identical Directory objects to add
        # concurrently
        directories = []
        entries = sample_data.directory7.to_dict()["entries"]
        for cnt in sample_data.contents:
            fentry = {
                "name": b"the_file",
                "type": "file",
                "target": cnt.sha1_git,
                "perms": from_disk.DentryPerms.content,
            }
            for entry in entries:
                directories.append(
                    Directory.from_dict({"entries": (entry,) + (fentry,)})
                )
            directories.append(Directory.from_dict({"entries": entries + (fentry,)}))
            directories.append(Directory.from_dict({"entries": (fentry,)}))
        directories *= 5

        # used to gather each thread execution result as triplets (thread_id,
        # result_type, payload), payload can be a normal result or an exception
        results = queue.Queue()

        def thread(dirs):
            try:
                with db_transaction(swh_storage) as (db, cur):
                    ret = swh_storage.directory_add(directories=dirs, db=db, cur=cur)
                    # give a chance for another thread to have some cpu
                    time.sleep(0.1)
                results.put((threading.get_ident(), "data", ret))
            except Exception as e:
                results.put((threading.get_ident(), "exc", e))

        ts = [threading.Thread(target=thread, args=([d],)) for d in directories]
        [t.start() for t in ts]
        [t.join() for t in ts]

        rs = [results.get(block=False) for _ in ts]

        with pytest.raises(queue.Empty):
            results.get(block=False)

        for tid, rtype, payload in rs:
            if rtype == "exc":
                # there might be UniqueViolation errors, but only on the
                # directory table, not the directory_entry_xxx ones
                assert (
                    payload.diag.constraint_name == "directory_pkey"
                ), "Got exception %r in Thread%s" % (payload, tid)

        for d in directories:
            entries = tuple(swh_storage.directory_get_entries(d.id).results)
            assert sorted(entries) == sorted(d.entries)


class TestPgStorage:
    """This class is dedicated for the rare case where the schema needs to
    be altered dynamically.

    Otherwise, the tests could be blocking when ran altogether.

    """

    def test_content_update_with_new_cols(self, swh_storage, sample_data):
        content, content2 = sample_data.contents[:2]

        swh_storage.journal_writer.journal = None  # TODO, not supported

        with db_transaction(swh_storage) as (_, cur):
            cur.execute(
                """alter table content
                           add column test text default null,
                           add column test2 text default null"""
            )

        swh_storage.content_add([content])

        cont = content.to_dict()
        cont["test"] = "value-1"
        cont["test2"] = "value-2"

        swh_storage.content_update([cont], keys=["test", "test2"])
        with db_transaction(swh_storage) as (_, cur):
            cur.execute(
                """SELECT sha1, sha1_git, sha256, length, status,
                   test, test2
                   FROM content WHERE sha1 = %s""",
                (cont["sha1"],),
            )

            datum = cur.fetchone()

        assert datum == (
            cont["sha1"],
            cont["sha1_git"],
            cont["sha256"],
            cont["length"],
            "visible",
            cont["test"],
            cont["test2"],
        )

        with db_transaction(swh_storage) as (_, cur):
            cur.execute(
                """alter table content drop column test,
                                               drop column test2"""
            )

    def test_content_add_db(self, swh_storage, sample_data):
        content = sample_data.content

        actual_result = swh_storage.content_add([content])

        assert actual_result == {
            "content:add": 1,
            "content:add:bytes": content.length,
        }

        if hasattr(swh_storage, "objstorage"):
            assert content.hashes() in swh_storage.objstorage.objstorage

        with db_transaction(swh_storage) as (_, cur):
            cur.execute(
                "SELECT sha1, sha1_git, sha256, length, status"
                " FROM content WHERE sha1 = %s",
                (content.sha1,),
            )
            datum = cur.fetchone()

        assert datum == (
            content.sha1,
            content.sha1_git,
            content.sha256,
            content.length,
            "visible",
        )

        contents = [
            obj
            for (obj_type, obj) in swh_storage.journal_writer.journal.objects
            if obj_type == "content"
        ]
        assert len(contents) == 1
        assert contents[0] == attr.evolve(content, data=None)

    def test_content_add_metadata_db(self, swh_storage, sample_data):
        content = attr.evolve(sample_data.content, data=None, ctime=now())

        actual_result = swh_storage.content_add_metadata([content])

        assert actual_result == {
            "content:add": 1,
        }

        if hasattr(swh_storage, "objstorage"):
            assert content.hashes() not in swh_storage.objstorage.objstorage
        with db_transaction(swh_storage) as (_, cur):
            cur.execute(
                "SELECT sha1, sha1_git, sha256, length, status"
                " FROM content WHERE sha1 = %s",
                (content.sha1,),
            )
            datum = cur.fetchone()
        assert datum == (
            content.sha1,
            content.sha1_git,
            content.sha256,
            content.length,
            "visible",
        )

        contents = [
            obj
            for (obj_type, obj) in swh_storage.journal_writer.journal.objects
            if obj_type == "content"
        ]
        assert len(contents) == 1
        assert contents[0] == content

    def test_skipped_content_add_db(self, swh_storage, sample_data):
        content, cont2 = sample_data.skipped_contents[:2]
        content2 = attr.evolve(cont2, blake2s256=None)

        actual_result = swh_storage.skipped_content_add([content, content, content2])

        assert 2 <= actual_result.pop("skipped_content:add") <= 3
        assert actual_result == {}

        with db_transaction(swh_storage) as (_, cur):
            cur.execute(
                "SELECT sha1, sha1_git, sha256, blake2s256, "
                "length, status, reason "
                "FROM skipped_content ORDER BY sha1_git"
            )

            dbdata = cur.fetchall()

        assert len(dbdata) == 2
        assert dbdata[0] == (
            content.sha1,
            content.sha1_git,
            content.sha256,
            content.blake2s256,
            content.length,
            "absent",
            "Content too long",
        )

        assert dbdata[1] == (
            content2.sha1,
            content2.sha1_git,
            content2.sha256,
            content2.blake2s256,
            content2.length,
            "absent",
            "Content too long",
        )

    def test_revision_get_displayname_behavior(self, swh_storage, sample_data):
        """Check revision_get behavior when displayname is set"""
        revision, revision2 = sample_data.revisions[:2]

        # Make authors and committers known
        revision = attr.evolve(
            revision,
            author=Person.from_fullname(b"author1 <author1@example.com>"),
            committer=Person.from_fullname(b"committer1 <committer1@example.com>"),
        )
        revision = attr.evolve(revision, id=revision.compute_hash())
        revision2 = attr.evolve(
            revision2,
            author=Person.from_fullname(b"author2 <author2@example.com>"),
            committer=Person.from_fullname(b"committer2 <committer2@example.com>"),
        )
        revision2 = attr.evolve(revision2, id=revision2.compute_hash())

        add_result = swh_storage.revision_add([revision, revision2])
        assert add_result == {"revision:add": 2}

        # Before displayname change
        revisions = swh_storage.revision_get([revision.id, revision2.id])
        assert revisions == [revision, revision2]

        displayname = b"Display Name <displayname@example.com>"

        with db_transaction(swh_storage) as (_, cur):
            cur.execute(
                "UPDATE person set displayname = %s where fullname = %s",
                (displayname, revision.author.fullname),
            )

        revisions = swh_storage.revision_get([revision.id, revision2.id])
        assert revisions == [
            attr.evolve(revision, author=Person.from_fullname(displayname)),
            revision2,
        ]

        revisions = swh_storage.revision_get(
            [revision.id, revision2.id], ignore_displayname=True
        )
        assert revisions == [revision, revision2]

    def test_revision_log_displayname_behavior(self, swh_storage, sample_data):
        """Check revision_log behavior when displayname is set"""
        revision, revision2 = sample_data.revisions[:2]

        # Make authors, committers and parenthood relationship known
        # (revision2 -[parent]-> revision1)
        revision = attr.evolve(
            revision,
            author=Person.from_fullname(b"author1 <author1@example.com>"),
            committer=Person.from_fullname(b"committer1 <committer1@example.com>"),
        )
        revision = attr.evolve(revision, id=revision.compute_hash())
        revision2 = attr.evolve(
            revision2,
            parents=(revision.id,),
            author=Person.from_fullname(b"author2 <author2@example.com>"),
            committer=Person.from_fullname(b"committer2 <committer2@example.com>"),
        )
        revision2 = attr.evolve(revision2, id=revision2.compute_hash())

        add_result = swh_storage.revision_add([revision, revision2])
        assert add_result == {"revision:add": 2}

        # Before displayname change
        revisions = swh_storage.revision_log([revision2.id])
        assert list(revisions) == [revision2.to_dict(), revision.to_dict()]

        displayname = b"Display Name <displayname@example.com>"

        with db_transaction(swh_storage) as (_, cur):
            cur.execute(
                "UPDATE person set displayname = %s where fullname = %s",
                (displayname, revision.author.fullname),
            )

        revisions = swh_storage.revision_log([revision2.id])
        assert list(revisions) == [
            revision2.to_dict(),
            attr.evolve(revision, author=Person.from_fullname(displayname)).to_dict(),
        ]

        revisions = swh_storage.revision_log([revision2.id], ignore_displayname=True)
        assert list(revisions) == [revision2.to_dict(), revision.to_dict()]

    def test_release_get_displayname_behavior(self, swh_storage, sample_data):
        """Check release_get behavior when displayname is set"""
        release, release2 = sample_data.releases[:2]

        # Make authors known
        release = attr.evolve(
            release,
            author=Person.from_fullname(b"author1 <author1@example.com>"),
        )
        release = attr.evolve(release, id=release.compute_hash())
        release2 = attr.evolve(
            release2,
            author=Person.from_fullname(b"author2 <author2@example.com>"),
        )
        release2 = attr.evolve(release2, id=release2.compute_hash())

        add_result = swh_storage.release_add([release, release2])
        assert add_result == {"release:add": 2}

        # Before displayname change
        releases = swh_storage.release_get([release.id, release2.id])
        assert releases == [release, release2]

        displayname = b"Display Name <displayname@example.com>"

        with db_transaction(swh_storage) as (_, cur):
            cur.execute(
                "UPDATE person set displayname = %s where fullname = %s",
                (displayname, release.author.fullname),
            )

        releases = swh_storage.release_get([release.id, release2.id])
        assert releases == [
            attr.evolve(release, author=Person.from_fullname(displayname)),
            release2,
        ]

        releases = swh_storage.release_get(
            [release.id, release2.id], ignore_displayname=True
        )
        assert releases == [release, release2]

    def test_clear_buffers(self, swh_storage):
        """Calling clear buffers on real storage does nothing"""
        assert swh_storage.clear_buffers() is None

    def test_flush(self, swh_storage):
        """Calling clear buffers on real storage does nothing"""
        assert swh_storage.flush() == {}

    def test_check_config(self, swh_storage):
        assert swh_storage.check_config(check_write=True)
        assert swh_storage.check_config(check_write=False)

    def test_check_config_dbversion(self, swh_storage, monkeypatch):
        swh_storage.current_version = -1
        assert swh_storage.check_config(check_write=True) is False
        assert swh_storage.check_config(check_write=False) is False



# Copyright (C) 2019-2023 The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import multiprocessing.util

from hypothesis import settings
import pytest

try:
    import pytest_cov
    import pytest_cov.embed
except ImportError:
    pytest_cov = None

from typing import Iterable

from swh.model.model import BaseContent, Origin
from swh.model.tests.generate_testdata import gen_contents, gen_origins
from swh.storage.interface import StorageInterface

# define tests profile. Full documentation is at:
# https://hypothesis.readthedocs.io/en/latest/settings.html#settings-profiles
settings.register_profile("fast", max_examples=5, deadline=5000)
settings.register_profile("slow", max_examples=20, deadline=5000)
# Load the fast profile by default to overcome default hypothesis values
# (max_examples=100, deadline=200) that are unsuitable for our tests.
# This can still be overloaded via the --hypothesis-profile option.
settings.load_profile("fast")

if pytest_cov is not None and int(pytest_cov.__version__.split(".")[0]) < 4:
    # pytest_cov + multiprocessing can cause a segmentation fault when starting
    # the child process <https://forge.softwareheritage.org/P706>; so we're
    # removing pytest-coverage's hook that runs when a child process starts.
    # This means code run in child processes won't be counted in the coverage
    # report, but this is not an issue because the only code that runs only in
    # child processes is the RPC server.
    registry = multiprocessing.util._afterfork_registry  # type: ignore[attr-defined]
    for key, value in registry.items():
        if value is pytest_cov.embed.multiprocessing_start:
            del registry[key]
            break
    else:
        assert False, "missing pytest_cov.embed.multiprocessing_start?"


@pytest.fixture
def swh_storage_backend_config(swh_storage_postgresql_backend_config):
    """Storage should test with its journal writer collaborator on"""
    yield {
        **swh_storage_postgresql_backend_config,
        "journal_writer": {
            "cls": "memory",
        },
    }


@pytest.fixture
def swh_contents(swh_storage: StorageInterface) -> Iterable[BaseContent]:
    contents = [BaseContent.from_dict(c) for c in gen_contents(n=20)]
    swh_storage.content_add([c for c in contents if c.status != "absent"])
    swh_storage.skipped_content_add([c for c in contents if c.status == "absent"])
    return contents


@pytest.fixture
def swh_origins(swh_storage: StorageInterface) -> Iterable[Origin]:
    origins = [Origin.from_dict(o) for o in gen_origins(n=100)]
    swh_storage.origin_add(origins)
    return origins



# Copyright (C) 2019-2020  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import datetime

from swh.model.model import Origin, OriginVisit, OriginVisitStatus
from swh.storage.algos.origin import (
    iter_origin_visit_statuses,
    iter_origin_visits,
    iter_origins,
    origin_get_latest_visit_status,
)
from swh.storage.interface import ListOrder
from swh.storage.tests.storage_tests import round_to_milliseconds
from swh.storage.utils import now


def test_iter_origins(swh_storage):
    origins = [
        Origin(url="bar"),
        Origin(url="qux"),
        Origin(url="quuz"),
    ]
    assert swh_storage.origin_add(origins) == {"origin:add": 3}

    # this returns all the origins, only the number of paged called is different
    assert list(iter_origins(swh_storage)) == origins
    assert list(iter_origins(swh_storage, limit=1)) == origins
    assert list(iter_origins(swh_storage, limit=2)) == origins


def test_origin_get_latest_visit_status_none(swh_storage, sample_data):
    """Looking up unknown objects should return nothing"""
    # unknown origin so no result
    assert origin_get_latest_visit_status(swh_storage, "unknown-origin") is None

    # unknown type so no result
    origin = sample_data.origin
    origin_visit = sample_data.origin_visit
    assert origin_visit.origin == origin.url

    swh_storage.origin_add([origin])
    swh_storage.origin_visit_add([origin_visit])[0]
    assert origin_visit.type != "unknown"
    actual_origin_visit = origin_get_latest_visit_status(
        swh_storage, origin.url, type="unknown"
    )
    assert actual_origin_visit is None

    actual_origin_visit = origin_get_latest_visit_status(
        swh_storage, origin.url, require_snapshot=True
    )
    assert actual_origin_visit is None


def init_storage_with_origin_visits(swh_storage, sample_data):
    """Initialize storage with origin/origin-visit/origin-visit-status"""
    snapshot = sample_data.snapshots[2]
    origin1, origin2 = sample_data.origins[:2]
    swh_storage.origin_add([origin1, origin2])

    ov1, ov2 = swh_storage.origin_visit_add(
        [
            OriginVisit(
                origin=origin1.url,
                date=sample_data.date_visit1,
                type=sample_data.type_visit1,
            ),
            OriginVisit(
                origin=origin2.url,
                date=sample_data.date_visit2,
                type=sample_data.type_visit2,
            ),
        ]
    )

    swh_storage.snapshot_add([snapshot])

    date_now = now()
    date_now = round_to_milliseconds(date_now)
    assert sample_data.date_visit1 < sample_data.date_visit2
    assert sample_data.date_visit2 < date_now

    # origin visit status 1 for origin visit 1
    ovs11 = OriginVisitStatus(
        origin=ov1.origin,
        visit=ov1.visit,
        date=ov1.date + datetime.timedelta(seconds=10),  # so it's not ignored
        type=ov1.type,
        status="partial",
        snapshot=None,
    )
    # origin visit status 2 for origin visit 1
    ovs12 = OriginVisitStatus(
        origin=ov1.origin,
        visit=ov1.visit,
        date=sample_data.date_visit2,
        type=ov1.type,
        status="ongoing",
        snapshot=None,
    )
    # origin visit status 1 for origin visit 2
    ovs21 = OriginVisitStatus(
        origin=ov2.origin,
        visit=ov2.visit,
        date=ov2.date + datetime.timedelta(seconds=10),  # so it's not ignored
        type=ov2.type,
        status="ongoing",
        snapshot=None,
    )
    # origin visit status 2 for origin visit 2
    ovs22 = OriginVisitStatus(
        origin=ov2.origin,
        visit=ov2.visit,
        date=date_now,
        type=ov2.type,
        status="full",
        snapshot=snapshot.id,
    )

    swh_storage.origin_visit_status_add([ovs11, ovs12, ovs21, ovs22])
    return {
        "origin": [origin1, origin2],
        "origin_visit": [ov1, ov2],
        "origin_visit_status": [ovs11, ovs12, ovs21, ovs22],
    }


def test_origin_get_latest_visit_status_filter_type(swh_storage, sample_data):
    """Filtering origin visit per types should yield consistent results"""
    objects = init_storage_with_origin_visits(swh_storage, sample_data)
    origin1, origin2 = objects["origin"]
    ov1, ov2 = objects["origin_visit"]
    ovs11, ovs12, _, ovs22 = objects["origin_visit_status"]

    # no visit for origin1 url with type_visit2
    assert (
        origin_get_latest_visit_status(
            swh_storage, origin1.url, type=sample_data.type_visit2
        )
        is None
    )

    # no visit for origin2 url with type_visit1
    assert (
        origin_get_latest_visit_status(
            swh_storage, origin2.url, type=sample_data.type_visit1
        )
        is None
    )

    # Two visits, both with no snapshot, take the most recent
    actual_ovs12 = origin_get_latest_visit_status(
        swh_storage, origin1.url, type=sample_data.type_visit1
    )
    assert isinstance(actual_ovs12, OriginVisitStatus)
    assert actual_ovs12 == ovs12
    assert actual_ovs12.origin == ov1.origin
    assert actual_ovs12.visit == ov1.visit
    assert actual_ovs12.type == sample_data.type_visit1

    # take the most recent visit with type_visit2
    actual_ovs22 = origin_get_latest_visit_status(
        swh_storage, origin2.url, type=sample_data.type_visit2
    )
    assert isinstance(actual_ovs22, OriginVisitStatus)
    assert actual_ovs22 == ovs22
    assert actual_ovs22.origin == ov2.origin
    assert actual_ovs22.visit == ov2.visit
    assert actual_ovs22.type == sample_data.type_visit2


def test_origin_get_latest_visit_status_filter_status(swh_storage, sample_data):
    objects = init_storage_with_origin_visits(swh_storage, sample_data)
    origin1, origin2 = objects["origin"]
    ov1, ov2 = objects["origin_visit"]
    ovs11, ovs12, _, ovs22 = objects["origin_visit_status"]

    # no partial status for that origin visit
    assert (
        origin_get_latest_visit_status(
            swh_storage, origin2.url, allowed_statuses=["partial"]
        )
        is None
    )

    # only 1 partial for that visit
    actual_ovs11 = origin_get_latest_visit_status(
        swh_storage, origin1.url, allowed_statuses=["partial"]
    )
    assert actual_ovs11 == ovs11
    assert actual_ovs11.origin == ov1.origin
    assert actual_ovs11.visit == ov1.visit
    assert actual_ovs11.type == sample_data.type_visit1

    # both status exist, take the latest one
    actual_ovs12 = origin_get_latest_visit_status(
        swh_storage, origin1.url, allowed_statuses=["partial", "ongoing"]
    )
    assert actual_ovs12 == ovs12
    assert actual_ovs12.origin == ov1.origin
    assert actual_ovs12.visit == ov1.visit
    assert actual_ovs12.type == sample_data.type_visit1

    assert isinstance(actual_ovs12, OriginVisitStatus)
    assert actual_ovs12 == ovs12
    assert actual_ovs12.origin == ov1.origin
    assert actual_ovs12.visit == ov1.visit
    assert actual_ovs12.type == sample_data.type_visit1

    # take the most recent visit with type_visit2
    actual_ovs22 = origin_get_latest_visit_status(
        swh_storage, origin2.url, allowed_statuses=["full"]
    )
    assert actual_ovs22 == ovs22
    assert actual_ovs22.origin == ov2.origin
    assert actual_ovs22.visit == ov2.visit
    assert actual_ovs22.type == sample_data.type_visit2


def test_origin_get_latest_visit_status_filter_snapshot(swh_storage, sample_data):
    objects = init_storage_with_origin_visits(swh_storage, sample_data)
    origin1, origin2 = objects["origin"]
    _, ov2 = objects["origin_visit"]
    _, _, _, ovs22 = objects["origin_visit_status"]

    # there is no visit with snapshot yet for that visit
    assert (
        origin_get_latest_visit_status(swh_storage, origin1.url, require_snapshot=True)
        is None
    )

    # visit status with partial status visit elected
    actual_ovs22 = origin_get_latest_visit_status(
        swh_storage, origin2.url, require_snapshot=True
    )
    assert actual_ovs22 == ovs22
    assert actual_ovs22.origin == ov2.origin
    assert actual_ovs22.visit == ov2.visit
    assert actual_ovs22.type == ov2.type

    date_now = now()

    # Add another visit
    swh_storage.origin_visit_add(
        [
            OriginVisit(
                origin=origin2.url,
                date=date_now,
                type=sample_data.type_visit2,
            ),
        ]
    )

    # Requiring the latest visit with a snapshot, we still find the previous visit
    ovs22 = origin_get_latest_visit_status(
        swh_storage, origin2.url, require_snapshot=True
    )
    assert actual_ovs22 == ovs22
    assert actual_ovs22.origin == ov2.origin
    assert actual_ovs22.visit == ov2.visit
    assert actual_ovs22.type == ov2.type


def test_iter_origin_visits(swh_storage, sample_data):
    """Iter over origin visits for an origin returns all visits"""
    origin1, origin2 = sample_data.origins[:2]
    swh_storage.origin_add([origin1, origin2])

    date_past = now() - datetime.timedelta(weeks=20)

    new_visits = []
    for visit_id in range(20):
        new_visits.append(
            OriginVisit(
                origin=origin1.url,
                date=date_past + datetime.timedelta(days=visit_id),
                type="git",
            )
        )

    visits = swh_storage.origin_visit_add(new_visits)
    reversed_visits = list(reversed(visits))

    # no limit, order asc
    actual_visits = list(iter_origin_visits(swh_storage, origin1.url))
    assert actual_visits == visits

    # no limit, order desc
    actual_visits = list(
        iter_origin_visits(swh_storage, origin1.url, order=ListOrder.DESC)
    )
    assert actual_visits == reversed_visits

    # no result
    actual_visits = list(iter_origin_visits(swh_storage, origin2.url))
    assert actual_visits == []


def test_iter_origin_visit_status(swh_storage, sample_data):
    origin1, origin2 = sample_data.origins[:2]
    swh_storage.origin_add([origin1])

    ov1 = swh_storage.origin_visit_add([sample_data.origin_visit])[0]
    assert ov1.origin == origin1.url

    date_past = now() - datetime.timedelta(weeks=20)

    ovs1 = OriginVisitStatus(
        origin=ov1.origin,
        visit=ov1.visit,
        date=ov1.date,
        type=ov1.type,
        status="created",
        snapshot=None,
    )
    new_visit_statuses = [ovs1]
    for i in range(20):
        status_date = date_past + datetime.timedelta(days=i)

        new_visit_statuses.append(
            OriginVisitStatus(
                origin=ov1.origin,
                visit=ov1.visit,
                date=status_date,
                type=ov1.type,
                status="created",
                snapshot=None,
            )
        )

    swh_storage.origin_visit_status_add(new_visit_statuses)
    reversed_visit_statuses = list(reversed(new_visit_statuses))

    # order asc
    actual_visit_statuses = list(
        iter_origin_visit_statuses(swh_storage, ov1.origin, ov1.visit)
    )
    assert actual_visit_statuses == new_visit_statuses

    # order desc
    actual_visit_statuses = list(
        iter_origin_visit_statuses(
            swh_storage, ov1.origin, ov1.visit, order=ListOrder.DESC
        )
    )
    assert actual_visit_statuses == reversed_visit_statuses

    # no result
    actual_visit_statuses = list(
        iter_origin_visit_statuses(swh_storage, origin2.url, ov1.visit)
    )
    assert actual_visit_statuses == []



# Copyright (C) 2020  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

# flake8: noqa
# because of long lines

import copy
import datetime
import json
import urllib.error

import attr

from swh.model.model import (
    MetadataAuthority,
    MetadataAuthorityType,
    MetadataFetcher,
    Origin,
    OriginVisit,
    OriginVisitStatus,
    RawExtrinsicMetadata,
    Snapshot,
    SnapshotBranch,
    SnapshotTargetType,
)
from swh.model.swhids import CoreSWHID, ExtendedObjectType, ExtendedSWHID
from swh.storage import get_storage
from swh.storage.interface import PagedResult
from swh.storage.migrate_extrinsic_metadata import (
    handle_row,
    pypi_origin_from_filename,
    pypi_project_from_filename,
)

FETCHER = MetadataFetcher(
    name="migrate-extrinsic-metadata-from-revisions",
    version="0.0.1",
)
PYPI_AUTHORITY = MetadataAuthority(
    type=MetadataAuthorityType.FORGE,
    url="https://pypi.org/",
)
SWH_AUTHORITY = MetadataAuthority(
    type=MetadataAuthorityType.REGISTRY,
    url="https://softwareheritage.org/",
)

DIRECTORY_ID = b"a" * 20
DIRECTORY_SWHID = ExtendedSWHID(
    object_type=ExtendedObjectType.DIRECTORY, object_id=DIRECTORY_ID
)


def now():
    return datetime.datetime.now(tz=datetime.timezone.utc)


def test_pypi_project_from_filename():
    files = [
        ("django-agent-trust-0.1.8.tar.gz", "django-agent-trust"),
        ("python_test-1.0.1.zip", "python_test"),
        ("py-evm-0.2.0a9.tar.gz", "py-evm"),
        ("collective.texttospeech-1.0rc1.tar.gz", "collective.texttospeech"),
        ("flatland-fork-0.4.post1.dev40550160.zip", "flatland-fork"),
        ("fake-factory-0.5.6-proper.tar.gz", "fake-factory"),
        ("ariane_procos-0.1.2-b05.tar.gz", "ariane_procos"),
        ("Yelpy-0.2.2dev.tar.gz", "Yelpy"),
        ("geventhttpclient_c-1.0a-t1.tar.gz", "geventhttpclient_c"),
        ("codeforlife-portal-1.0.0.post.dev618.tar.gz", "codeforlife-portal"),
        ("ChecklistDSL-0.0.1.alpha.1.tar.gz", "ChecklistDSL"),
        ("transifex-1.1.0beta.tar.gz", "transifex"),
        ("thespian-2.5.10.tar.bz2", "thespian"),
        ("janis pipelines-0.5.3.tar.gz", "janis-pipelines"),
        ("pants-1.0.0-beta.2.tar.gz", "pants"),
        ("uforge_python_sdk-3.8.4-RC15.tar.gz", "uforge_python_sdk"),
        ("virtuoso-0.11.0.48.b5865c2b46fb.tar.gz", "virtuoso"),
        ("cloud_ftp-v1.0.0.tar.gz", "cloud_ftp"),
        ("frozenordereddict-1.0.0.tgz", "frozenordereddict"),
        ("pywebsite-0.1.2pre.tar.gz", "pywebsite"),
        ("Flask Unchained-0.2.0.tar.gz", "Flask-Unchained"),
        ("mongomotor-0.13.0.n.tar.gz", "mongomotor"),
        ("datahaven-rev8784.tar.gz", "datahaven"),
        ("geopandas-0.1.0.dev-120d5ee.tar.gz", "geopandas"),
        ("aimmo-v0.1.1-alpha.post.dev61.tar.gz", "aimmo"),
        ("django-migrations-plus-0.1.0.dev5.gdd1abd3.tar.gz", "django-migrations-plus"),
        ("function_shield.tar.gz", "function_shield"),
        ("Dtls-0.1.0.sdist_with_openssl.mingw-win32.tar.gz", "Dtls"),
        ("pytz-2005m.tar.gz", "pytz"),
        ("python-librsync-0.1-3.tar.gz", "python-librsync"),
        ("powny-1.4.0-alpha-20141205-1452-f5a2b03.tar.gz", "powny"),
        ("stp-3pc-batch-0.1.11.tar.gz", "stp-3pc-batch"),
        ("obedient.powny-3.0.0-alpha-20141027-2102-9e53ebd.tar.gz", "obedient.powny"),
        ("mojimoji-0.0.9_2.tar.gz", "mojimoji"),
        ("devpi-theme-16-2.0.0.tar.gz", "devpi-theme-16"),
        ("Orange3-WONDER-1-1.0.7.tar.gz", "Orange3-WONDER-1"),
        ("obj-34.tar.gz", "obj"),
        ("pytorch-ignite-nightly-20190825.tar.gz", "pytorch-ignite-nightly"),
        ("tlds-2019081900.tar.gz", "tlds"),
        ("dominator-12.1.2-alpha-20141027-1446-ad46e0f.tar.gz", "dominator"),
        ("waferslim-1.0.0-py3.1.zip", "waferslim"),
        ("Beaver-21.tar.gz", "Beaver"),
        ("aimmo-0.post.dev460.tar.gz", "aimmo"),
        ("ohai-1!0.tar.gz", "ohai"),
        ("nevolution-risk-139.tar.gz", "nevolution-risk"),
        ("collective.topicitemsevent-0.1dvl.tar.gz", "collective.topicitemsevent"),
        ("lesscpy-0.9g.tar.gz", "lesscpy"),
        ("SpiNNStorageHandlers-1!4.0.0a1.tar.gz", "SpiNNStorageHandlers"),
        ("limnoria-2013-03-27T16:32:26+0100.tar.gz", "limnoria"),
        (
            "sPyNNakerExternalDevicesPlugin-1!4.0.0a2.tar.gz",
            "sPyNNakerExternalDevicesPlugin",
        ),
        ("django-bootstrap-italia_0.1.tar.gz", "django-bootstrap-italia"),
        ("sPyNNaker8-1!4.0.0a1.tar.gz", "sPyNNaker8"),
        ("betahaus.openmember-0.1adev-r1651.tar.gz", "betahaus.openmember"),
        ("mailer.0.8.0.zip", "mailer"),
        ("pytz-2005k.tar.bz2", "pytz"),
        ("aha.plugin.microne-0.62bdev.tar.gz", "aha.plugin.microne"),
        ("youtube_dl_server-alpha.3.tar.gz", "youtube_dl_server"),
        ("json-extensions-b76bc7d.tar.gz", "json-extensions"),
        ("LitReview-0.6989ev.tar.gz", "LitReview"),
        ("django_options-r5.tar.gz", "django_options"),
        ("ddlib-2013-11-07.tar.gz", "ddlib"),
        ("python-morfeusz-0.3000+py3k.tar.gz", "python-morfeusz"),
        ("gaepytz-2011h.zip", "gaepytz"),
        ("ftldat-r3.tar.gz", "ftldat"),
        ("tigretoolbox-0.0.0-py2.7-linux-x86_64.egg", None),
        (
            "Greater than, equal, or less Library-0.1.tar.gz",
            "Greater-than-equal-or-less-Library",
        ),
        ("upstart--main-.-VLazy.object.at.0x104ba8b50-.tar.gz", "upstart"),
        ("duckduckpy0.1.tar.gz", "duckduckpy"),
        ("QUI for MPlayer snapshot_9-14-2011.zip", "QUI-for-MPlayer"),
        ("Eddy's Memory Game-1.0.zip", "Eddy-s-Memory-Game"),
        ("jekyll2nikola-0-0-1.tar.gz", "jekyll2nikola"),
        ("ore.workflowed-0-6-2.tar.gz", "ore.workflowed"),
        ("instancemanager-1.0rc-r34317.tar.gz", "instancemanager"),
        ("OrzMC_W&L-1.0.0.tar.gz", "OrzMC-W-L"),
        ("use0mk.tar.gz", "use0mk"),
        ("play-0-develop-1-gd67cd85.tar.gz", "play"),
        ("mosaic-nist-2.0b1+f98ae80.tar.gz", "mosaic-nist"),
        ("pypops-201408-r3.tar.gz", "pypops"),
    ]

    for filename, project in files:
        assert pypi_project_from_filename(filename) == project


def test_pypi_origin_from_project_name(mocker):
    origin_url = "https://pypi.org/project/ProjectName/"

    storage = get_storage("memory")

    revision_id = b"41" * 10
    snapshot_id = b"42" * 10
    storage.origin_add([Origin(url=origin_url)])
    storage.origin_visit_add(
        [OriginVisit(origin=origin_url, visit=1, date=now(), type="pypi")]
    )
    storage.origin_visit_status_add(
        [
            OriginVisitStatus(
                origin=origin_url,
                visit=1,
                date=now(),
                status="partial",
                snapshot=snapshot_id,
            )
        ]
    )
    storage.snapshot_add(
        [
            Snapshot(
                id=snapshot_id,
                branches={
                    b"foo": SnapshotBranch(
                        target_type=SnapshotTargetType.REVISION,
                        target=revision_id,
                    )
                },
            )
        ]
    )

    class response:
        code = 200

        def read(self):
            return b'{"info": {"name": "ProjectName"}}'

    mock_urlopen = mocker.patch(
        "swh.storage.migrate_extrinsic_metadata.urlopen",
        return_value=response(),
    )

    assert (
        pypi_origin_from_filename(storage, revision_id, "ProjectName-1.0.0.tar.gz")
        == origin_url
    )
    mock_urlopen.assert_not_called()
    assert (
        pypi_origin_from_filename(storage, revision_id, "projectname-1.0.0.tar.gz")
        == origin_url
    )
    mock_urlopen.assert_called_once_with("https://pypi.org/pypi/projectname/json/")


def test_pypi_1():
    """Tests loading a revision generated by a new PyPI loader that
    has a provider."""

    extrinsic_metadata = {
        "url": "https://files.pythonhosted.org/packages/70/89/a498245baf1bf3dde73d3da00b4b067a8aa7c7378ad83472078803ea3e43/m3-ui-2.2.73.tar.gz",
        "size": 3933168,
        "digests": {
            "md5": "a374ac3f655e97df5db5335e2142d344",
            "sha256": "1bc2756f7d0d2e15cf5880ca697682ff35e8b58116bf73eb9c78b3db358c5b7d",
        },
        "has_sig": False,
        "filename": "m3-ui-2.2.73.tar.gz",
        "downloads": -1,
        "md5_digest": "a374ac3f655e97df5db5335e2142d344",
        "packagetype": "sdist",
        "upload_time": "2019-11-11T06:21:20",
        "comment_text": "",
        "python_version": "source",
        "requires_python": None,
        "upload_time_iso_8601": "2019-11-11T06:21:20.073082Z",
    }

    original_artifacts = [
        {
            "length": 3933168,
            "filename": "m3-ui-2.2.73.tar.gz",
            "checksums": {
                "sha1": "9f4ec7ce64b7fea4b122e85d47ea31146c367b03",
                "sha256": "1bc2756f7d0d2e15cf5880ca697682ff35e8b58116bf73eb9c78b3db358c5b7d",
            },
        }
    ]

    row = {
        "id": b"\x00\x00\x07a{S\xe7\xb1E\x8fi]\xd0}\xe4\xceU\xaf\x15\x17",
        "directory": DIRECTORY_ID,
        "date": datetime.datetime(
            2019,
            11,
            11,
            6,
            21,
            20,
            tzinfo=datetime.timezone.utc,
        ),
        "committer_date": datetime.datetime(
            2019,
            11,
            11,
            6,
            21,
            20,
            tzinfo=datetime.timezone.utc,
        ),
        "type": "tar",
        "message": b"2.2.73",
        "metadata": {
            "extrinsic": {
                "raw": extrinsic_metadata,
                "when": "2020-01-23T18:43:09.109407+00:00",
                "provider": "https://pypi.org/pypi/m3-ui/json",
            },
            "intrinsic": {
                "raw": {
                    "name": "m3-ui",
                    "summary": "======",
                    "version": "2.2.73",
                    # ...
                    "metadata_version": "1.1",
                },
                "tool": "PKG-INFO",
            },
            "original_artifact": original_artifacts,
        },
    }

    origin_url = "https://pypi.org/project/m3-ui/"

    storage = get_storage("memory")
    storage.origin_add([Origin(url=origin_url)])
    storage.metadata_authority_add(
        [
            attr.evolve(PYPI_AUTHORITY, metadata={}),
            attr.evolve(SWH_AUTHORITY, metadata={}),
        ]
    )
    storage.metadata_fetcher_add([FETCHER])

    deposit_cur = None
    handle_row(copy.deepcopy(row), storage, deposit_cur, dry_run=False)

    revision_swhid = CoreSWHID.from_string(
        "swh:1:rev:000007617b53e7b1458f695dd07de4ce55af1517"
    )
    assert storage.raw_extrinsic_metadata_get(
        DIRECTORY_SWHID,
        authority=PYPI_AUTHORITY,
    ) == PagedResult(
        results=[
            RawExtrinsicMetadata(
                target=DIRECTORY_SWHID,
                discovery_date=datetime.datetime(
                    2020,
                    1,
                    23,
                    18,
                    43,
                    9,
                    109407,
                    tzinfo=datetime.timezone.utc,
                ),
                authority=PYPI_AUTHORITY,
                fetcher=FETCHER,
                format="pypi-project-json",
                metadata=json.dumps(extrinsic_metadata).encode(),
                origin=origin_url,
                revision=revision_swhid,
            ),
        ],
        next_page_token=None,
    )
    assert storage.raw_extrinsic_metadata_get(
        DIRECTORY_SWHID,
        authority=SWH_AUTHORITY,
    ) == PagedResult(
        results=[
            RawExtrinsicMetadata(
                target=DIRECTORY_SWHID,
                discovery_date=datetime.datetime(
                    2020,
                    1,
                    23,
                    18,
                    43,
                    9,
                    109407,
                    tzinfo=datetime.timezone.utc,
                ),
                authority=SWH_AUTHORITY,
                fetcher=FETCHER,
                format="original-artifacts-json",
                metadata=json.dumps(original_artifacts).encode(),
                origin=origin_url,
                revision=revision_swhid,
            ),
        ],
        next_page_token=None,
    )


def test_pypi_2(mocker):
    """Tests loading a revision generated by an old PyPI loader that
    does not have a provider, but has 'project' metadata."""

    mocker.patch(
        "swh.storage.migrate_extrinsic_metadata.urlopen",
        side_effect=urllib.error.HTTPError(None, 404, "Not Found", None, None),
    )

    extrinsic_metadata = {
        "name": "jupyterhub-simx",
        "author": "Jupyter Development Team",
        "license": "BSD",
        "summary": "JupyterHub: A multi-user server for Jupyter notebooks",
        "version": "1.0.5",
        # ...
    }

    source_original_artifacts = [
        {
            "url": "https://files.pythonhosted.org/packages/72/28/a8098763d78e2c4607cb67602c0d726a97ac38d4c1f531aac28f49de2e1a/jupyterhub-simx-1.0.5.tar.gz",
            "date": "2019-01-23T22:10:55",
            "sha1": "ede3eadd5a06e70912e3ba7cfccef789c4ad3168",
            "size": 2346538,
            "sha256": "0399d7f5f0d90c525d369f0507ad0e8ef8729c1c7fa63aadfc46a27514d14a46",
            "filename": "jupyterhub-simx-1.0.5.tar.gz",
            "sha1_git": "734301124712182eb30fc90e97cc18cef5432f02",
            "blake2s256": "bb4aa82ffb5891a05dcf6d4dce3ad56fd2c18e9abdba9d20972910649d869322",
            "archive_type": "tar",
        }
    ]

    dest_original_artifacts = [
        {
            "url": "https://files.pythonhosted.org/packages/72/28/a8098763d78e2c4607cb67602c0d726a97ac38d4c1f531aac28f49de2e1a/jupyterhub-simx-1.0.5.tar.gz",
            "filename": "jupyterhub-simx-1.0.5.tar.gz",
            "archive_type": "tar",
            "length": 2346538,
            "checksums": {
                "sha1": "ede3eadd5a06e70912e3ba7cfccef789c4ad3168",
                "sha256": "0399d7f5f0d90c525d369f0507ad0e8ef8729c1c7fa63aadfc46a27514d14a46",
                "sha1_git": "734301124712182eb30fc90e97cc18cef5432f02",
                "blake2s256": "bb4aa82ffb5891a05dcf6d4dce3ad56fd2c18e9abdba9d20972910649d869322",
            },
        }
    ]

    row = {
        "id": b"\x00\x00\x04\xd68,J\xd4\xc0Q\x92fbl6U\x1f\x0eQ\xca",
        "directory": DIRECTORY_ID,
        "date": datetime.datetime(
            2019, 1, 23, 22, 10, 55, tzinfo=datetime.timezone.utc
        ),
        "committer_date": datetime.datetime(
            2019, 1, 23, 22, 10, 55, tzinfo=datetime.timezone.utc
        ),
        "type": "tar",
        "message": b"1.0.5",
        "metadata": {
            "project": extrinsic_metadata,
            "original_artifact": source_original_artifacts,
        },
    }

    origin_url = "https://pypi.org/project/jupyterhub-simx/"

    storage = get_storage("memory")

    storage.origin_add([Origin(url=origin_url)])
    storage.metadata_authority_add(
        [
            attr.evolve(PYPI_AUTHORITY, metadata={}),
            attr.evolve(SWH_AUTHORITY, metadata={}),
        ]
    )
    storage.metadata_fetcher_add([FETCHER])
    deposit_cur = None

    handle_row(copy.deepcopy(row), storage, deposit_cur, dry_run=False)

    revision_swhid = CoreSWHID.from_string(
        "swh:1:rev:000004d6382c4ad4c0519266626c36551f0e51ca"
    )
    assert storage.raw_extrinsic_metadata_get(
        DIRECTORY_SWHID,
        authority=PYPI_AUTHORITY,
    ) == PagedResult(
        results=[
            RawExtrinsicMetadata(
                target=DIRECTORY_SWHID,
                discovery_date=datetime.datetime(
                    2019,
                    1,
                    23,
                    22,
                    10,
                    55,
                    tzinfo=datetime.timezone.utc,
                ),
                authority=PYPI_AUTHORITY,
                fetcher=FETCHER,
                format="pypi-project-json",
                metadata=json.dumps(extrinsic_metadata).encode(),
                origin=None,
                revision=revision_swhid,
            ),
        ],
        next_page_token=None,
    )
    assert storage.raw_extrinsic_metadata_get(
        DIRECTORY_SWHID,
        authority=SWH_AUTHORITY,
    ) == PagedResult(
        results=[
            RawExtrinsicMetadata(
                target=DIRECTORY_SWHID,
                discovery_date=datetime.datetime(
                    2019,
                    1,
                    23,
                    22,
                    10,
                    55,
                    tzinfo=datetime.timezone.utc,
                ),
                authority=SWH_AUTHORITY,
                fetcher=FETCHER,
                format="original-artifacts-json",
                metadata=json.dumps(dest_original_artifacts).encode(),
                origin=None,
                revision=revision_swhid,
            ),
        ],
        next_page_token=None,
    )


def test_pypi_3(mocker):
    """Tests loading a revision generated by a very old PyPI loader that
    does not have a provider or has 'project' metadata."""

    mocker.patch(
        "swh.storage.migrate_extrinsic_metadata.urlopen",
        side_effect=urllib.error.HTTPError(None, 404, "Not Found", None, None),
    )

    source_original_artifact = {
        "url": "https://files.pythonhosted.org/packages/34/4f/30087f22eaae8ad7077a28ce157342745a2977e264b8a8e4e7f804a8aa5e/PyPDFLite-0.1.32.tar.gz",
        "date": "2014-05-07T22:03:00",
        "sha1": "3289269f75b4111dd00eaea53e00330db9a1db12",
        "size": 46644,
        "sha256": "911497d655cf7ef6530c5b57773dad7da97e21cf4d608ad9ad1e38bd7bec7824",
        "filename": "PyPDFLite-0.1.32.tar.gz",
        "sha1_git": "1e5c38014731242cfa8594839bcba8a0c4e158c5",
        "blake2s256": "45792e57873f56d385c694e36c98a580cbba60d5ea91eb6fd0a2d1c71c1fb385",
        "archive_type": "tar",
    }

    dest_original_artifacts = [
        {
            "url": "https://files.pythonhosted.org/packages/34/4f/30087f22eaae8ad7077a28ce157342745a2977e264b8a8e4e7f804a8aa5e/PyPDFLite-0.1.32.tar.gz",
            "filename": "PyPDFLite-0.1.32.tar.gz",
            "archive_type": "tar",
            "length": 46644,
            "checksums": {
                "sha1": "3289269f75b4111dd00eaea53e00330db9a1db12",
                "sha256": "911497d655cf7ef6530c5b57773dad7da97e21cf4d608ad9ad1e38bd7bec7824",
                "sha1_git": "1e5c38014731242cfa8594839bcba8a0c4e158c5",
                "blake2s256": "45792e57873f56d385c694e36c98a580cbba60d5ea91eb6fd0a2d1c71c1fb385",
            },
        }
    ]

    row = {
        "id": b"N\xa9\x91|\xdfS\xcd\x13SJ\x04.N\xb3x{\x86\xc84\xd2",
        "directory": DIRECTORY_ID,
        "date": datetime.datetime(2014, 5, 7, 22, 3, tzinfo=datetime.timezone.utc),
        "committer_date": datetime.datetime(
            2014, 5, 7, 22, 3, tzinfo=datetime.timezone.utc
        ),
        "type": "tar",
        "message": b"0.1.32",
        "metadata": {"original_artifact": source_original_artifact},
    }

    origin_url = "https://pypi.org/project/PyPDFLite/"

    storage = get_storage("memory")

    storage.origin_add([Origin(url=origin_url)])
    storage.metadata_authority_add(
        [
            attr.evolve(PYPI_AUTHORITY, metadata={}),
            attr.evolve(SWH_AUTHORITY, metadata={}),
        ]
    )
    storage.metadata_fetcher_add([FETCHER])
    deposit_cur = None
    handle_row(copy.deepcopy(row), storage, deposit_cur, dry_run=False)

    revision_swhid = CoreSWHID.from_string(
        "swh:1:rev:4ea9917cdf53cd13534a042e4eb3787b86c834d2"
    )

    assert storage.raw_extrinsic_metadata_get(
        DIRECTORY_SWHID,
        authority=PYPI_AUTHORITY,
    ) == PagedResult(
        results=[],
        next_page_token=None,
    )
    assert storage.raw_extrinsic_metadata_get(
        DIRECTORY_SWHID,
        authority=SWH_AUTHORITY,
    ) == PagedResult(
        results=[
            RawExtrinsicMetadata(
                target=DIRECTORY_SWHID,
                discovery_date=datetime.datetime(
                    2014,
                    5,
                    7,
                    22,
                    3,
                    tzinfo=datetime.timezone.utc,
                ),
                authority=SWH_AUTHORITY,
                fetcher=FETCHER,
                format="original-artifacts-json",
                metadata=json.dumps(dest_original_artifacts).encode(),
                origin=None,
                revision=revision_swhid,
            ),
        ],
        next_page_token=None,
    )


def test_pypi_good_origin():
    """Tests loading a revision whose origin we can find"""

    source_original_artifact = {
        "url": "https://files.pythonhosted.org/packages/34/4f/30087f22eaae8ad7077a28ce157342745a2977e264b8a8e4e7f804a8aa5e/PyPDFLite-0.1.32.tar.gz",
        "date": "2014-05-07T22:03:00",
        "sha1": "3289269f75b4111dd00eaea53e00330db9a1db12",
        "size": 46644,
        "sha256": "911497d655cf7ef6530c5b57773dad7da97e21cf4d608ad9ad1e38bd7bec7824",
        "filename": "PyPDFLite-0.1.32.tar.gz",
        "sha1_git": "1e5c38014731242cfa8594839bcba8a0c4e158c5",
        "blake2s256": "45792e57873f56d385c694e36c98a580cbba60d5ea91eb6fd0a2d1c71c1fb385",
        "archive_type": "tar",
    }

    dest_original_artifacts = [
        {
            "url": "https://files.pythonhosted.org/packages/34/4f/30087f22eaae8ad7077a28ce157342745a2977e264b8a8e4e7f804a8aa5e/PyPDFLite-0.1.32.tar.gz",
            "filename": "PyPDFLite-0.1.32.tar.gz",
            "archive_type": "tar",
            "length": 46644,
            "checksums": {
                "sha1": "3289269f75b4111dd00eaea53e00330db9a1db12",
                "sha256": "911497d655cf7ef6530c5b57773dad7da97e21cf4d608ad9ad1e38bd7bec7824",
                "sha1_git": "1e5c38014731242cfa8594839bcba8a0c4e158c5",
                "blake2s256": "45792e57873f56d385c694e36c98a580cbba60d5ea91eb6fd0a2d1c71c1fb385",
            },
        }
    ]

    revision_id = b"N\xa9\x91|\xdfS\xcd\x13SJ\x04.N\xb3x{\x86\xc84\xd2"
    row = {
        "id": revision_id,
        "directory": DIRECTORY_ID,
        "date": datetime.datetime(2014, 5, 7, 22, 3, tzinfo=datetime.timezone.utc),
        "committer_date": datetime.datetime(
            2014, 5, 7, 22, 3, tzinfo=datetime.timezone.utc
        ),
        "type": "tar",
        "message": b"0.1.32",
        "metadata": {"original_artifact": source_original_artifact},
    }

    origin_url = "https://pypi.org/project/PyPDFLite/"

    storage = get_storage("memory")

    snapshot_id = b"42" * 10
    storage.origin_add([Origin(url=origin_url)])
    storage.origin_visit_add(
        [OriginVisit(origin=origin_url, visit=1, date=now(), type="pypi")]
    )
    storage.origin_visit_status_add(
        [
            OriginVisitStatus(
                origin=origin_url,
                visit=1,
                date=now(),
                status="partial",
                snapshot=snapshot_id,
            )
        ]
    )
    storage.snapshot_add(
        [
            Snapshot(
                id=snapshot_id,
                branches={
                    b"foo": SnapshotBranch(
                        target_type=SnapshotTargetType.REVISION,
                        target=revision_id,
                    )
                },
            )
        ]
    )
    storage.metadata_authority_add(
        [
            attr.evolve(PYPI_AUTHORITY, metadata={}),
            attr.evolve(SWH_AUTHORITY, metadata={}),
        ]
    )
    storage.metadata_fetcher_add([FETCHER])
    deposit_cur = None
    handle_row(copy.deepcopy(row), storage, deposit_cur, dry_run=False)

    revision_swhid = CoreSWHID.from_string(
        "swh:1:rev:4ea9917cdf53cd13534a042e4eb3787b86c834d2"
    )

    assert storage.raw_extrinsic_metadata_get(
        DIRECTORY_SWHID,
        authority=PYPI_AUTHORITY,
    ) == PagedResult(
        results=[],
        next_page_token=None,
    )
    assert storage.raw_extrinsic_metadata_get(
        DIRECTORY_SWHID,
        authority=SWH_AUTHORITY,
    ) == PagedResult(
        results=[
            RawExtrinsicMetadata(
                target=DIRECTORY_SWHID,
                discovery_date=datetime.datetime(
                    2014,
                    5,
                    7,
                    22,
                    3,
                    tzinfo=datetime.timezone.utc,
                ),
                authority=SWH_AUTHORITY,
                fetcher=FETCHER,
                format="original-artifacts-json",
                metadata=json.dumps(dest_original_artifacts).encode(),
                origin=origin_url,
                revision=revision_swhid,
            ),
        ],
        next_page_token=None,
    )



# Copyright (C) 2020-2025  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import functools
import logging

from cassandra import InvalidRequest
import pytest

from swh.model.model import Origin
from swh.storage import get_storage
from swh.storage.cassandra.migrations import MIGRATIONS, Migration

logger = logging.getLogger(__name__)


@pytest.fixture
def swh_storage_backend_config(swh_storage_cassandra_backend_config):
    return swh_storage_cassandra_backend_config


@pytest.fixture
def invoke(swh_storage_cassandra_backend_config):
    from .test_cli import invoke

    @functools.wraps(invoke)
    def newf(*args, **kwargs):
        assert "local_config" not in kwargs
        return invoke(
            *args,
            local_config={"storage": swh_storage_cassandra_backend_config},
            **kwargs,
        )

    return newf


@pytest.mark.cassandra
class TestCassandraCli:
    def test_init(
        self,
        swh_storage_cassandra_backend_config,
    ):
        # not using the the invoke fixture because it uses the keyspace that is already
        # initialized, which would make this test pointless
        from .test_cli import invoke

        result = invoke(
            "cassandra",
            "init",
            local_config={"storage": swh_storage_cassandra_backend_config},
        )
        assert result.exit_code == 0, result.output
        assert result.output == "Done.\n"

        # Check we can write and read to it
        storage = get_storage(**swh_storage_cassandra_backend_config)
        origin = Origin(url="http://example.org")
        storage.origin_add([origin])
        assert storage.origin_get([origin.url]) == [origin]

    def test_list_migrations(self, swh_storage, invoke, mocker):
        # keep only the first migration
        mocker.patch("swh.storage.cassandra.migrations.MIGRATIONS", MIGRATIONS[0:1])

        result = invoke("cassandra", "list-migrations")
        assert result.exit_code == 0, result.output
        assert result.output == (
            "2024-12-12_init (required): completed\n"
            "    Dummy migration that represents the database schema as of v2.9.0\n"
            "\n"
        )

    def test_list_migrations_pending(self, swh_storage, invoke, mocker):
        new_migration = Migration(
            id="2024-12-19_test_migration",
            dependencies=set(),
            min_read_version="2.9.0",
            script=None,
            help="Test migration",
            required=False,
        )
        mocker.patch(
            "swh.storage.cassandra.migrations.MIGRATIONS",
            [MIGRATIONS[0], new_migration],
        )

        result = invoke("cassandra", "list-migrations")
        assert result.exit_code == 0, result.output
        assert result.output == (
            "2024-12-12_init (required): completed\n"
            "    Dummy migration that represents the database schema as of v2.9.0\n"
            "\n"
            "2024-12-19_test_migration: pending\n"
            "    Test migration\n"
            "\n"
        )

    def test_upgrade_all(self, swh_storage, invoke, mocker):
        def migration_script(cql_runner):
            cql_runner.execute_with_retries(f"USE {cql_runner.keyspace}", [])
            cql_runner.execute_with_retries(
                "CREATE TABLE IF NOT EXISTS test_table (abc blob PRIMARY KEY);", []
            )

        new_migration = Migration(
            id="2024-12-19_test_migration",
            dependencies=set(),
            min_read_version="2.9.0",
            script=migration_script,
            help="Test migration",
            required=False,
        )
        mocker.patch(
            "swh.storage.cassandra.migrations.MIGRATIONS", [*MIGRATIONS, new_migration]
        )

        try:
            # create the table
            result = invoke("cassandra", "upgrade")
            assert result.exit_code == 0, result.output
            assert result.output == "Done.\n"

            # check the table exists
            swh_storage._cql_runner.execute_with_retries(
                f"SELECT * FROM {swh_storage.keyspace}.test_table", []
            )
        finally:
            swh_storage._cql_runner.execute_with_retries(
                f"DROP TABLE {swh_storage.keyspace}.test_table", []
            )

    def test_upgrade_all_from_v2_9(self, swh_storage, invoke, mocker):
        """Tests upgrading from v2.9.x, which did not have a 'migrations' table."""
        assert len(MIGRATIONS) == 1, (
            "This test won't work correctly after we make more changes to the schema, "
            "as it relies on the schema being v2.9's plus only the migrations table."
        )
        cql_runner = swh_storage._cql_runner

        cql_runner.execute_with_retries(
            f"DROP TABLE {cql_runner.keyspace}.migration", []
        )

        def migration_script(cql_runner):
            cql_runner.execute_with_retries(f"USE {cql_runner.keyspace}", [])
            cql_runner.execute_with_retries(
                "CREATE TABLE IF NOT EXISTS test_table (abc blob PRIMARY KEY);", []
            )

        new_migration = Migration(
            id="2024-12-19_test_migration",
            dependencies=set(),
            min_read_version="2.9.0",
            script=migration_script,
            help="Test migration",
            required=False,
        )
        mocker.patch(
            "swh.storage.cassandra.migrations.MIGRATIONS", [*MIGRATIONS, new_migration]
        )

        try:
            # create the table
            result = invoke("cassandra", "upgrade")
            assert result.exit_code == 0, result.output
            assert result.output == "Done.\n"

            # check the table exists
            swh_storage._cql_runner.execute_with_retries(
                f"SELECT * FROM {swh_storage.keyspace}.test_table", []
            )
        finally:
            swh_storage._cql_runner.execute_with_retries(
                f"DROP TABLE {swh_storage.keyspace}.test_table", []
            )

    def test_upgrade_crashing(self, swh_storage, invoke, mocker):
        class TestException(Exception):
            pass

        def migration_script(cql_runner):
            raise TestException("Oh no, a crash!")

        new_migration = Migration(
            id="2024-12-19_test_migration",
            dependencies=set(),
            min_read_version="2.9.0",
            script=migration_script,
            help="Test migration",
            required=False,
        )
        mocker.patch(
            "swh.storage.cassandra.migrations.MIGRATIONS",
            [MIGRATIONS[0], new_migration],
        )

        # run the buggy migration
        result = invoke("cassandra", "upgrade")
        assert result.exit_code == 1, result.output
        assert result.output == ""  # raised an exception, traceback on stderr

        result = invoke("cassandra", "list-migrations")
        assert result.exit_code == 0, result.output
        assert result.output == (
            "2024-12-12_init (required): completed\n"
            "    Dummy migration that represents the database schema as of v2.9.0\n"
            "\n"
            "2024-12-19_test_migration: running\n"
            "    Test migration\n"
            "\n"
        )

    def test_upgrade_partial(self, swh_storage, invoke, mocker):
        def create_test_table_script(cql_runner):
            cql_runner.execute_with_retries(f"USE {cql_runner.keyspace}", [])
            cql_runner.execute_with_retries(
                "CREATE TABLE IF NOT EXISTS test_table (abc blob PRIMARY KEY);", []
            )

        def drop_test_table_script(cql_runner):
            cql_runner.execute_with_retries(f"USE {cql_runner.keyspace}", [])
            cql_runner.execute_with_retries("DROP TABLE test_table;", [])

        new_migration1 = Migration(
            id="2024-12-19_create_test_table",
            dependencies=set(),
            min_read_version="2.9.0",
            script=create_test_table_script,
            help="Test migration",
            required=False,
        )
        new_migration2 = Migration(
            id="2024-12-19_drop_test_table",
            dependencies={"2024-12-19_create_test_table"},
            min_read_version="2.9.0",
            script=drop_test_table_script,
            help="Test migration",
            required=False,
        )
        mocker.patch(
            "swh.storage.cassandra.migrations.MIGRATIONS",
            [*MIGRATIONS, new_migration1, new_migration2],
        )

        # create the table
        result = invoke(
            "cassandra", "upgrade", "--migration", "2024-12-19_create_test_table"
        )
        assert result.exit_code == 0, result.output
        assert result.output == "Done.\n"

        # check the table exists
        swh_storage._cql_runner.execute_with_retries(
            f"SELECT * FROM {swh_storage.keyspace}.test_table", []
        )

        # drop the table
        result = invoke(
            "cassandra", "upgrade", "--migration", "2024-12-19_drop_test_table"
        )
        assert result.exit_code == 0, result.output
        assert result.output == "Done.\n"

        # check the table does not exist anymore
        with pytest.raises(InvalidRequest):
            swh_storage._cql_runner.execute_with_retries(
                f"SELECT * FROM {swh_storage.keyspace}.test_table", []
            )

    @pytest.mark.parametrize("required", [True, False])
    def test_upgrade_manual(self, swh_storage, invoke, mocker, required):
        """Tries to apply a migration that cannot run automatically"""
        new_migration1 = Migration(
            id="2025-03-18_manual1",
            dependencies=set(),
            min_read_version="2.9.0",
            script=None,
            help="Test migration",
            required=False,
        )
        new_migration2 = Migration(
            id="2025-03-18_manual2",
            dependencies=set(),
            min_read_version="2.9.0",
            script=None,
            help="Test migration",
            required=required,
        )
        mocker.patch(
            "swh.storage.cassandra.migrations.MIGRATIONS",
            [*MIGRATIONS, new_migration1, new_migration2],
        )

        # run migration before its dependency
        result = invoke(
            "cassandra",
            "upgrade",
            "--migration",
            "2025-03-18_manual1",
            "--migration",
            "2025-03-18_manual2",
        )
        if required:
            assert result.exit_code == 4, result.output
            assert result.output == (
                "Some migrations need to be manually applied: "
                "2025-03-18_manual1, 2025-03-18_manual2\n"
                "Including these required migrations: "
                "2025-03-18_manual2\n"
            )
        else:
            assert result.exit_code == 5, result.output
            assert result.output == (
                "Some migrations need to be manually applied: "
                "2025-03-18_manual1, 2025-03-18_manual2\n"
            )

        # check the migration was not marked as applied
        result = invoke("cassandra", "list-migrations")
        assert result.exit_code == 0, result.output
        assert result.output == (
            "2024-12-12_init (required): completed\n"
            "    Dummy migration that represents the database schema as of v2.9.0\n"
            "\n"
            "2025-03-18_manual1: running\n"
            "    Test migration\n"
            "\n"
            f"2025-03-18_manual2{' (required)' if required else ''}: running\n"
            "    Test migration\n"
            "\n"
        )

        # check the tables still do not exist
        with pytest.raises(InvalidRequest):
            swh_storage._cql_runner.execute_with_retries(
                f"SELECT * FROM {swh_storage.keyspace}.test_table1", []
            )
        with pytest.raises(InvalidRequest):
            swh_storage._cql_runner.execute_with_retries(
                f"SELECT * FROM {swh_storage.keyspace}.test_table2", []
            )

    @pytest.mark.parametrize(
        "migration1_manual,migration2_required",
        [
            pytest.param(True, True, id="migration1=manual,migration2=required"),
            pytest.param(True, False, id="migration1=manual,migration2=optional"),
            pytest.param(False, True, id="migration1=scripted,migration2=required"),
            pytest.param(False, False, id="migration1=scripted,migration2=optional"),
        ],
    )
    def test_upgrade_disordered(
        self,
        swh_storage,
        invoke,
        mocker,
        migration1_manual,
        migration2_required,
    ):
        """Tries to apply a migration before its dependency"""

        def create_test_table1_script(cql_runner):
            cql_runner.execute_with_retries(f"USE {cql_runner.keyspace}", [])
            cql_runner.execute_with_retries(
                "CREATE TABLE IF NOT EXISTS test_table1 (abc blob PRIMARY KEY);", []
            )

        def create_test_table2_script(cql_runner):
            cql_runner.execute_with_retries(f"USE {cql_runner.keyspace}", [])
            cql_runner.execute_with_retries(
                "CREATE TABLE IF NOT EXISTS test_table1 (abc blob PRIMARY KEY);", []
            )

        new_migration1 = Migration(
            id="2024-12-19_create_test_table1",
            dependencies=set(),
            min_read_version="2.9.0",
            script=None if migration1_manual else create_test_table1_script,
            help="Test migration",
            required=False,
        )
        new_migration2 = Migration(
            id="2024-12-19_create_test_table2",
            dependencies={"2024-12-19_create_test_table1"},
            min_read_version="2.9.0",
            script=create_test_table2_script,
            help="Test migration",
            required=migration2_required,
        )
        mocker.patch(
            "swh.storage.cassandra.migrations.MIGRATIONS",
            [*MIGRATIONS, new_migration1, new_migration2],
        )

        # run migration before its dependency
        result = invoke(
            "cassandra", "upgrade", "--migration", "2024-12-19_create_test_table2"
        )
        if migration2_required:
            assert result.exit_code == 6, result.output
            assert result.output == (
                "Some migrations could not be applied because a dependency is missing: "
                "2024-12-19_create_test_table2\n"
                "Including these required migrations: "
                "2024-12-19_create_test_table2\n"
            )
        else:
            assert result.exit_code == 7, result.output
            assert result.output == (
                "Some migrations could not be applied because a dependency is missing: "
                "2024-12-19_create_test_table2\n"
            )

        # check the migration was not marked as applied
        result = invoke("cassandra", "list-migrations")
        assert result.exit_code == 0, result.output
        assert result.output == (
            "2024-12-12_init (required): completed\n"
            "    Dummy migration that represents the database schema as of v2.9.0\n"
            "\n"
            "2024-12-19_create_test_table1: pending\n"
            "    Test migration\n"
            "\n"
            "2024-12-19_create_test_table2"
            f"{' (required)' if migration2_required else ''}: pending\n"
            "    Test migration\n"
            "\n"
        )

        # check the tables still do not exist
        with pytest.raises(InvalidRequest):
            swh_storage._cql_runner.execute_with_retries(
                f"SELECT * FROM {swh_storage.keyspace}.test_table1", []
            )
        with pytest.raises(InvalidRequest):
            swh_storage._cql_runner.execute_with_retries(
                f"SELECT * FROM {swh_storage.keyspace}.test_table2", []
            )

    def test_mark_upgraded(self, swh_storage, invoke, mocker):
        def create_test_table1_script(cql_runner):
            cql_runner.execute_with_retries(f"USE {cql_runner.keyspace}", [])
            cql_runner.execute_with_retries(
                "CREATE TABLE IF NOT EXISTS test_table1 (abc blob PRIMARY KEY);", []
            )

        def create_test_table2_script(cql_runner):
            cql_runner.execute_with_retries(f"USE {cql_runner.keyspace}", [])
            cql_runner.execute_with_retries(
                "CREATE TABLE IF NOT EXISTS test_table2 (abc blob PRIMARY KEY);", []
            )

        new_migration1 = Migration(
            id="2024-12-19_create_test_table1",
            dependencies=set(),
            min_read_version="2.9.0",
            script=create_test_table1_script,
            help="Test migration",
            required=False,
        )
        new_migration2 = Migration(
            id="2024-12-19_create_test_table2",
            dependencies={"2024-12-19_create_test_table1"},
            min_read_version="2.9.0",
            script=create_test_table2_script,
            help="Test migration",
            required=False,
        )
        mocker.patch(
            "swh.storage.cassandra.migrations.MIGRATIONS",
            [*MIGRATIONS, new_migration1, new_migration2],
        )
        # Pretend the first migration was applied
        result = invoke(
            "cassandra", "mark-upgraded", "--migration", "2024-12-19_create_test_table1"
        )
        assert result.exit_code == 0, result.output
        assert (
            result.output
            == "Migrations 2024-12-19_create_test_table1 marked as complete.\n"
        )

        # run dependent migration
        try:
            result = invoke(
                "cassandra", "upgrade", "--migration", "2024-12-19_create_test_table2"
            )
            assert result.exit_code == 0, result.output
            assert result.output == "Done.\n"

            # check the first table still does not exist
            with pytest.raises(InvalidRequest):
                swh_storage._cql_runner.execute_with_retries(
                    f"SELECT * FROM {swh_storage.keyspace}.test_table1", []
                )

            # check the second table now exists
            swh_storage._cql_runner.execute_with_retries(
                f"SELECT * FROM {swh_storage.keyspace}.test_table2", []
            )
        finally:
            try:
                swh_storage._cql_runner.execute_with_retries(
                    f"DROP TABLE {swh_storage.keyspace}.test_table2", []
                )
            except BaseException:
                pass



# Copyright (C) 2019-2023 The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

from functools import partial
import logging
from typing import Dict, Iterable, List, Literal, Mapping, Sequence, Tuple, cast
import warnings

from swh.core.utils import grouper
from swh.model.model import (
    BaseModel,
    Content,
    Directory,
    RawExtrinsicMetadata,
    Release,
    Revision,
    SkippedContent,
)
from swh.storage import get_storage
from swh.storage.interface import StorageInterface

logger = logging.getLogger(__name__)

LObjectType = Literal[
    "raw_extrinsic_metadata",
    "content",
    "content_metadata",
    "skipped_content",
    "directory",
    "revision",
    "release",
    "snapshot",
    "extid",
]

# In the order objects are flushed
OBJECT_TYPES: Tuple[LObjectType, ...] = (
    "raw_extrinsic_metadata",
    "content",
    "content_metadata",
    "skipped_content",
    "directory",
    "revision",
    "release",
    "snapshot",
    "extid",
)

DEFAULT_BUFFER_THRESHOLDS: Dict[str, int] = {
    "content": 10000,
    "content_metadata": 10000,
    "content_bytes": 100 * 1024 * 1024,
    "skipped_content": 10000,
    "directory": 25000,
    "directory_entries": 200000,
    "raw_extrinsic_metadata": 1000,
    "revision": 100000,
    "revision_parents": 200000,
    "revision_bytes": 100 * 1024 * 1024,
    "release": 100000,
    "release_bytes": 100 * 1024 * 1024,
    "snapshot": 25000,
    "extid": 10000,
}


def estimate_revision_size(revision: Revision) -> int:
    """Estimate the size of a revision, by summing the size of variable length fields"""
    s = 20 * len(revision.parents)

    if revision.message:
        s += len(revision.message)

    if revision.author is not None:
        s += len(revision.author.fullname)
    if revision.committer is not None:
        s += len(revision.committer.fullname)
    s += sum(len(h) + len(v) for h, v in revision.extra_headers)

    return s


def estimate_release_size(release: Release) -> int:
    """Estimate the size of a release, by summing the size of variable length fields"""
    s = 0
    if release.message:
        s += len(release.message)
    if release.author:
        s += len(release.author.fullname)

    return s


class BufferingProxyStorage:
    """
    Storage implementation in charge of accumulating objects prior to
    discussing with the "main" storage.
    When the number of objects of any given type exceeds the configure threshold,
    then objects of this type (and of other types, see the 'Flush order' below) are
    flushed to the backend.

    Deduplicates values based on a tuple of keys depending on the object type.

    Sample configuration use case for buffering storage:

    .. code-block:: yaml

        storage:
          cls: buffer
          args:
            storage:
              cls: remote
              args: http://storage.internal.staging.swh.network:5002/
            min_batch_size:
              content: 10000
              content_bytes: 100000000
              skipped_content: 10000
              directory: 5000
              directory_entries: 100000
              raw_extrinsic_metadata: 1000
              revision: 1000
              revision_parents: 2000
              revision_bytes: 100000000
              release: 10000
              release_bytes: 100000000
              snapshot: 5000

    Flush order
    ===========

    In order not to create holes when the process crashes (eg. by adding a revision
    but crashing before adding its root directory), objects are always flushed in
    reverse topological order of their types:

    {flush_order}

    This guarantees not to create holes between objects of different types.

    However, holes may still be created between objects of the same type when using
    a backend storage which inserts neither sequentially not transactionally,
    such as the :mod:`Cassandra <swh.storage.cassandra>` backend or the
    :mod:`tenacious <swh.storage.proxies.tenacious>` proxy, so this is mostly best
    effort, and relies on both of these inserting objects **mostly** in sequential
    order.
    """.format(
        flush_order="\n    ".join("#. " + object_type for object_type in OBJECT_TYPES)
    )

    def __init__(self, storage: Mapping, min_batch_size: Mapping = {}):
        self.storage: StorageInterface = get_storage(**storage)

        self._buffer_thresholds = {**DEFAULT_BUFFER_THRESHOLDS, **min_batch_size}

        self._objects: Dict[LObjectType, Dict[Tuple[str, ...], BaseModel]] = {
            k: {} for k in OBJECT_TYPES
        }
        self._sizes: Dict[LObjectType, int] = {k: 0 for k in OBJECT_TYPES}
        self._directory_entries: int = 0
        self._revision_parents: int = 0

    def __del__(self):
        if not hasattr(self, "_sizes"):
            return

        leftovers = False
        for object_type in OBJECT_TYPES:
            if self._sizes[object_type] != 0:
                leftovers = True
                break

        if not leftovers:
            return

        # Attempts to flush nonetheless to minimize data loss
        flush_failed = False
        try:
            self.flush()
        except Exception:
            flush_failed = True

        warnings.warn(
            "Some objects were still present in the memory of the buffering "
            "proxy during shutdown. "
            f"{'**They are now probably lost!** ' if flush_failed else ''}"
            "A call to `storage.flush()` is probably missing."
        )

    def __getattr__(self, key: str):
        if key.endswith("_add"):
            object_type = key.rsplit("_", 1)[0]
            if object_type in OBJECT_TYPES:
                return partial(
                    self.object_add,
                    object_type=cast(LObjectType, object_type),
                    keys=["id"],
                )
        if key == "storage":
            raise AttributeError(key)
        return getattr(self.storage, key)

    def content_add(self, contents: Sequence[Content]) -> Dict[str, int]:
        """Push contents to write to the storage in the buffer.

        Following policies apply:

        - if the buffer's threshold is hit, flush content to the storage.
        - otherwise, if the total size of buffered contents's threshold is hit,
          flush content to the storage.

        """
        stats = self.object_add(
            contents,
            object_type="content",
            keys=["sha1", "sha1_git", "sha256", "blake2s256"],
        )
        if not stats:
            # We did not flush based on number of objects; check total size
            self._sizes["content"] += sum(c.length for c in contents)
            if self._sizes["content"] >= self._buffer_thresholds["content_bytes"]:
                return self.flush(["raw_extrinsic_metadata", "content"])

        return stats

    def content_add_metadata(self, contents: Sequence[Content]) -> Dict[str, int]:
        """Push content metadata (content without the payload) to write to the
        storage in the buffer.
        """
        stats = self.object_add(
            contents,
            object_type="content_metadata",
            keys=["sha1", "sha1_git", "sha256", "blake2s256"],
        )
        return stats

    def skipped_content_add(self, contents: Sequence[SkippedContent]) -> Dict[str, int]:
        return self.object_add(
            contents,
            object_type="skipped_content",
            keys=["sha1", "sha1_git", "sha256", "blake2s256"],
        )

    def directory_add(self, directories: Sequence[Directory]) -> Dict[str, int]:
        stats = self.object_add(directories, object_type="directory", keys=["id"])

        if not stats:
            # We did not flush based on number of objects; check the number of entries
            self._directory_entries += sum(len(d.entries) for d in directories)
            if self._directory_entries >= self._buffer_thresholds["directory_entries"]:
                return self.flush(["raw_extrinsic_metadata", "content", "directory"])

        return stats

    def revision_add(self, revisions: Sequence[Revision]) -> Dict[str, int]:
        stats = self.object_add(revisions, object_type="revision", keys=["id"])

        if not stats:
            # We did not flush based on number of objects; check the number of
            # parents and estimated size
            self._revision_parents += sum(len(r.parents) for r in revisions)
            self._sizes["revision"] += sum(estimate_revision_size(r) for r in revisions)
            if (
                self._revision_parents >= self._buffer_thresholds["revision_parents"]
                or self._sizes["revision"] >= self._buffer_thresholds["revision_bytes"]
            ):
                return self.flush(
                    ["raw_extrinsic_metadata", "content", "directory", "revision"]
                )

        return stats

    def release_add(self, releases: Sequence[Release]) -> Dict[str, int]:
        stats = self.object_add(releases, object_type="release", keys=["id"])

        if not stats:
            # We did not flush based on number of objects; check the estimated size
            self._sizes["release"] += sum(estimate_release_size(r) for r in releases)
            if self._sizes["release"] >= self._buffer_thresholds["release_bytes"]:
                return self.flush(
                    [
                        "raw_extrinsic_metadata",
                        "content",
                        "directory",
                        "revision",
                        "release",
                    ]
                )

        return stats

    def object_add(
        self,
        objects: Sequence[BaseModel],
        *,
        object_type: LObjectType,
        keys: Iterable[str],
    ) -> Dict[str, int]:
        """Push objects to write to the storage in the buffer. Flushes the
        buffer to the storage if the threshold is hit.

        """
        buffer_ = self._objects[object_type]
        for obj in objects:
            obj_key = tuple(getattr(obj, key) for key in keys)
            buffer_[obj_key] = obj
        if len(buffer_) >= self._buffer_thresholds[object_type]:
            return self.flush()

        return {}

    def flush(
        self, object_types: Sequence[LObjectType] = OBJECT_TYPES
    ) -> Dict[str, int]:
        summary: Dict[str, int] = {}

        def update_summary(stats):
            for k, v in stats.items():
                summary[k] = v + summary.get(k, 0)

        for object_type in object_types:
            buffer_ = self._objects[object_type]
            if not buffer_:
                continue

            if logger.isEnabledFor(logging.DEBUG):
                log = "Flushing %s objects of type %s"
                log_args = [len(buffer_), object_type]

                if object_type == "content":
                    log += " (%s bytes)"
                    log_args.append(
                        sum(cast(Content, c).length for c in buffer_.values())
                    )

                elif object_type == "directory":
                    log += " (%s entries)"
                    log_args.append(
                        sum(len(cast(Directory, d).entries) for d in buffer_.values())
                    )

                elif object_type == "revision":
                    log += " (%s parents, %s estimated bytes)"
                    log_args.extend(
                        (
                            sum(
                                len(cast(Revision, r).parents) for r in buffer_.values()
                            ),
                            sum(
                                estimate_revision_size(cast(Revision, r))
                                for r in buffer_.values()
                            ),
                        )
                    )

                elif object_type == "release":
                    log += " (%s estimated bytes)"
                    log_args.append(
                        sum(
                            estimate_release_size(cast(Release, r))
                            for r in buffer_.values()
                        )
                    )

                elif object_type == "raw_extrinsic_metadata":
                    log += " (%s estimated bytes)"
                    log_args.append(
                        sum(
                            len(cast(RawExtrinsicMetadata, r).metadata)
                            for r in buffer_.values()
                        )
                    )

                logger.debug(log, *log_args)

            batches = grouper(buffer_.values(), n=self._buffer_thresholds[object_type])
            for batch in batches:
                objs = list(batch)
                if object_type == "content_metadata":
                    stats = self.storage.content_add_metadata(cast(List[Content], objs))
                else:
                    add_fn = getattr(self.storage, "%s_add" % object_type)
                    stats = add_fn(objs)
                update_summary(stats)

        # Flush underlying storage
        stats = self.storage.flush(object_types)
        update_summary(stats)

        self.clear_buffers(object_types)

        return summary

    def clear_buffers(self, object_types: Sequence[LObjectType] = OBJECT_TYPES) -> None:
        """Clear objects from current buffer.

        WARNING:

            data that has not been flushed to storage will be lost when this
            method is called. This should only be called when `flush` fails and
            you want to continue your processing.

        """
        for object_type in object_types:
            buffer_ = self._objects[object_type]
            buffer_.clear()
            self._sizes[object_type] = 0
            if object_type == "directory":
                self._directory_entries = 0
            elif object_type == "revision":
                self._revision_parents = 0

        self.storage.clear_buffers(object_types)
